
from datamart_isi.joiners.join_feature.feature_classes import *
from functools import reduce
import numpy as np


class FeatureFactory:
    subclasses = {
        (DistributeType.CATEGORICAL, DataType.NUMBER): CategoricalNumberFeature,
        (DistributeType.CATEGORICAL, DataType.STRING): CategoricalStringFeature,
        (DistributeType.TOKEN_CATEGORICAL, DataType.STRING): CategoricalTokenFeature,
        (DistributeType.NON_CATEGORICAL, DataType.NUMBER): NonCategoricalNumberFeature,
        (DistributeType.NON_CATEGORICAL, DataType.STRING): NonCategoricalStringFeature
    }

    @classmethod
    def create(cls, df: pd.DataFrame, indexes, df_metadata):
        """
        TODO: dynamically generate subclass of FeatureBase, by profiled info, datatype etc.

        """
        # set default values:
        metadata = cls._get_feature_metadata(df_metadata, indexes) or {}
        data_type = None
        distribute_type = DistributeType.NON_CATEGORICAL

        if len(indexes) > 1:
            distribute_type = DistributeType.TOKEN_CATEGORICAL
            if cls._try_pd_to_datetime(df, indexes):
                data_type = DataType.DATETIME
        else:
            # single column, not datetime
            idx = indexes[0]
            profiles = metadata.get('dsbox_profiled', {})

            if len(df.iloc[:, idx]) // len(df.iloc[:, idx].unique()) >= 1.5:
                distribute_type = DistributeType.CATEGORICAL
            elif profiles:
                most_common_tokens = profiles.get('most_common_tokens')
                if most_common_tokens and cls._get_greater_than(most_common_tokens) >= len(most_common_tokens)//2:
                    distribute_type = DistributeType.TOKEN_CATEGORICAL

            dtype = df.iloc[:, idx].dtype
            if dtype == np.int64 or dtype == np.float64:
                data_type = DataType.NUMBER
            else:
                semantic_types = metadata.get('semantic_type')
                profiles = metadata.get('dsbox_profiled', {})
                data_type = cls._get_data_type_by_semantic_type(semantic_types) \
                            or cls._get_data_type_by_profile(profiles)
                if not data_type and cls._try_pd_to_datetime(df, indexes):
                    data_type = DataType.DATETIME

        return cls.get_instance(df, indexes, metadata, data_type or DataType.STRING, distribute_type)

    @classmethod
    def get_instance(cls, df, indices, metadata, data_type, distribute_type):
        constructor = cls.get_constructor(data_type, distribute_type)
        return constructor(df, indices, metadata, distribute_type, data_type)

    @classmethod
    def get_constructor(cls, data_type, distribute_type=None):
        if data_type == DataType.DATETIME:
            return DatetimeFeature
        return cls.subclasses.get((distribute_type, data_type))

    @staticmethod
    def _get_feature_metadata(metadata, indices):
        if metadata.get('variables') and indices and indices[0] < len(metadata.get('variables')):
            return metadata['variables'][indices[0]]

    @staticmethod
    def _get_avg(list_of_dict, key='count'):
        if len(list_of_dict):
            return sum([_.get(key) for _ in list_of_dict])/len(list_of_dict)

    @staticmethod
    def _get_greater_than(list_of_dict, key='count', threshold=2, inclusive=True):
        if inclusive:
            return reduce(lambda x, y: x + 1 if float(y[key]) >= threshold else x, list_of_dict, 0)
        return reduce(lambda x, y: x + 1 if float(y[key]) > threshold else x, list_of_dict, 0)

    @staticmethod
    def _get_data_type_by_semantic_type(semantic_types: list):
        # TODO: it would be better if we have a close set of used semantic_type, \
        # and map them to either STRING, NUMBER or DATETIME
        if semantic_types and len(semantic_types):
            unique_types = set(t.rsplit('/', 1)[-1].lower() for t in semantic_types)
            if 'time' in unique_types or 'date' in unique_types or 'datetime' in unique_types:
                return DataType.DATETIME
            if 'float' in unique_types or 'int' in unique_types or 'number' in unique_types:
                return DataType.NUMBER

    @staticmethod
    def _get_data_type_by_profile(profiles):
        numeric_ratio = profiles.get('ratio_of_numeric_values')
        if numeric_ratio and numeric_ratio >= 0.99:
            return DataType.NUMBER

    @staticmethod
    def _try_pd_to_datetime(df, indices):
        try:
            if len(indices) == 1:
                _ = pd.to_datetime(df.iloc[[0, len(df) - 1], indices[0]])
            else:
                _ = pd.to_datetime(df.iloc[[0, len(df)-1], indices])
            return True
        except ValueError:
            return False





import warnings
import dateutil.parser
# from datamart_isi.materializers.materializer_base import MaterializerBase
import importlib
import os
import sys
import json
from jsonschema import validate
from termcolor import colored
import typing
import pandas as pd
import tempfile
from datetime import datetime
from datamart_isi.utilities.timeout import timeout
import re
import wikifier
import traceback
# from datamart_isi.utilities.caching import Cache, EntryState
from io import StringIO
from ast import literal_eval

class Utils:
    DEFAULT_DESCRIPTION = {
        "materialization": {
            "python_path": "default_materializer"
        },
        "variables": []
    }

    @staticmethod
    def materialize(metadata) -> pd.DataFrame:
        if 'dataset' not in metadata:
            raise ValueError("No dataset name found")
        else:
            dataset_url = metadata['url']['value']
        from dsbox.datapreprocessing.cleaner.data_profile import Profiler, Hyperparams as ProfilerHyperparams
        from dsbox.datapreprocessing.cleaner.cleaning_featurizer import CleaningFeaturizer, CleaningFeaturizerHyperparameter
        file_type = metadata.get("file_type") or ""
        if file_type == "":
            # no file type get, try to guess
            file_type = dataset_url.split(".")[-1]
        else:
            file_type = file_type['value']

        if file_type == "wikitable":
            extra_information = literal_eval(metadata['extra_information']['value'])
            loaded_data = Utils.materialize_for_wikitable(dataset_url, file_type, extra_information)
        else:
            loaded_data = Utils.materialize_for_general(dataset_url, file_type)

        # run dsbox's profiler and cleaner
        hyper1 = ProfilerHyperparams.defaults()
        profiler = Profiler(hyperparams=hyper1)
        profiled_df = profiler.produce(inputs=loaded_data).value
        hyper2 = CleaningFeaturizerHyperparameter.defaults()
        clean_f = CleaningFeaturizer(hyperparams=hyper2)
        clean_f.set_training_data(inputs=profiled_df)
        clean_f.fit()
        cleaned_df = pd.DataFrame(clean_f.produce(inputs=profiled_df).value)
        wikifier_res = wikifier.produce(cleaned_df)

        return wikifier_res

    @staticmethod
    def materialize_for_wikitable(dataset_url:str, file_type:str, extra_information:str) -> pd.DataFrame:
        from datamart_isi.materializers.wikitables_materializer import WikitablesMaterializer
        materializer = WikitablesMaterializer()
        loaded_data = materializer.get_one(dataset_url, extra_information['xpath'])
        return loaded_data

    @staticmethod
    def materialize_for_general(dataset_url:str, file_type:str) -> pd.DataFrame:
        from datamart_isi.materializers.general_materializer import GeneralMaterializer
        general_materializer = GeneralMaterializer()
        file_metadata = {
            "materialization": {
                "arguments": {
                    "url": dataset_url,
                    # one example here: "url": "http://insight.dev.schoolwires.com/HelpAssets/C2Assets/C2Files/C2ImportFamRelSample.csv",
                    "file_type": file_type
                }
            }
        }

        try:
            result = general_materializer.get(metadata=file_metadata).to_csv(index=False)
            # remove last \n so that we will not get an extra useless row
            if result[-1] == "\n":
                result = result[:-1]

            loaded_data = StringIO(result)
            loaded_data = pd.read_csv(loaded_data,dtype="str")
            return loaded_data
        except:
            traceback.print_exc()
            raise ValueError("Materializing from " + dataset_url + " failed!")

    @staticmethod
    def calculate_dsbox_features(data: pd.DataFrame, metadata: typing.Union[dict, None],
                                 selected_columns: typing.Set[int] = None) -> dict:
        """Calculate dsbox features, add to metadata dictionary

         Args:
             data: dataset as a pandas dataframe
             metadata: metadata dict

         Returns:
              updated metadata dict
         """

        from datamart_isi.profilers.dsbox_profiler import DSboxProfiler
        if not metadata:
            return metadata
        return DSboxProfiler().profile(inputs=data, metadata=metadata, selected_columns=selected_columns)

    @classmethod
    def generate_metadata_from_dataframe(cls, data: pd.DataFrame, original_meta: dict=None) -> dict:
        """Generate a default metadata just from the data, without the dataset schema

         Args:
             data: pandas DataFrame

         Returns:
              metadata dict
         """
        from datamart_isi.profilers.basic_profiler import BasicProfiler, VariableMetadata, GlobalMetadata

        global_metadata = GlobalMetadata.construct_global(description=cls.DEFAULT_DESCRIPTION)
        for col_offset in range(data.shape[1]):
            variable_metadata = BasicProfiler.basic_profiling_column(
                description={},
                variable_metadata=VariableMetadata.construct_variable(description={}),
                column=data.iloc[:, col_offset]
            )
            global_metadata.add_variable_metadata(variable_metadata)
        global_metadata = BasicProfiler.basic_profiling_entire(global_metadata=global_metadata,
                                                               data=data)
        if original_meta:
            global_metadata.value.update(original_meta)
        return global_metadata.value


import pandas as pd
import os
import requests
import time
import json


class TwoRavensProfiler(object):
    def __init__(self):
        self.tr_url = 'http://metadata.2ravens.org/preprocess/api/process-single-file'
        pass

    def profile(self, inputs: pd.DataFrame, metadata: dict) -> dict:
        """Save metadata json to file.

        Args:
            inputs: pandas dataframe
            metadata: dict

        Returns:
            dict
        """

        temp_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'datamart_temp.csv')
        self.file_save(temp_path, inputs)
        f = dict(source_file=open(temp_path, 'rb'))
        raw_iter_times = 5
        callback_url = self.tr_raw_reader(f, raw_iter_times)
        if callback_url:
            time.sleep(1)
            json_iter_times = 5
            dataset, variables = self.tr_json_reader(callback_url, json_iter_times)
            if dataset or variables:
                metadata['two_ravens_dataset'] = dataset
                if 'variables' in metadata:
                    self.add_variables(variables, metadata['variables'])
        self.file_delete(temp_path)
        return metadata

    def add_variables(self, variables, metadata: list):
        for v in variables:
            for m in metadata:
                if v == m['name']:
                    m['two_ravens_variables'] = variables[v]

    def tr_raw_reader(self, file, raw_iter_times):
        if raw_iter_times > 0:
            raw_iter_times -= 1
            try:
                r = requests.post(self.tr_url, files=file, timeout=2)
                if r.status_code != 200:
                    print('Error: ', r.text)
                    time.sleep(1)
                    self.tr_raw_reader(file, raw_iter_times)
                else:
                    resp_json = r.json()
                    if resp_json['data']['state'] != 'FAILURE':
                        print('Raw json got!')
                        return resp_json['callback_url']
            except:
                print('Error')
                time.sleep(1)
                self.tr_raw_reader(file, raw_iter_times)
        else:
            return None

    def tr_json_reader(self, url, json_iter_times=5):
        if json_iter_times > 0:
            json_iter_times -= 1
            try:
                f = requests.get(url, timeout=3)
                if f.status_code == 200:
                    content_json = json.loads(f.text)
                    print('State: ', content_json['data']['state'])
                    if content_json['data']['state'] == 'SUCCESS':
                        dataset = content_json['data']['summary_metadata']['dataset']
                        variables = content_json['data']['summary_metadata']['variables']
                        print('Two Ravens data got!')
                        return dataset, variables
                    else:
                        time.sleep(1)
                        self.tr_json_reader(url, json_iter_times)
            except:
                print('Error')
                time.sleep(1)
                self.tr_json_reader(url, json_iter_times)
        else:
            return None, None

    def file_save(self, file_path: str, csv_file: pd.DataFrame):
        self.file_delete(file_path)
        csv_file.to_csv(file_path, sep=',', index=False)

    def file_delete(self, file_path: str):
        exist = os.path.isfile(file_path)
        if exist:
            os.remove(file_path)


import requests
import os, json, re, argparse
from etk.etk import ETK
from etk.extractors.excel_extractor import ExcelExtractor
from etk.knowledge_graph import KGSchema, URI, Literal, LiteralType, Subject, Reification
from etk.etk_module import ETKModule
from etk.wikidata.entity import WDProperty, WDItem
from etk.wikidata.value import Datatype, Item, TimeValue, Precision, QuantityValue
from etk.wikidata.statement import WDReference
from etk.wikidata import serialize_change_record
from SPARQLWrapper import SPARQLWrapper, JSON

# DATA_ABS_ADDRESS = "/Users/pszekely/Downloads/datamart_demo/wikidata-wikifier/wikifier/wikidata/data/"
# DATA_ABS_ADDRESS = "/Users/minazuki/Desktop/studies/master/2018Summer/DSBOX_2019/wikidata-wikifier/wikifier/wikidata/data/"
DATA_ABS_ADDRESS = "data/"


class FBI_Crime_Model():
    def __init__(self):

        self.value_dict = {
            'value': '$col,$row',
            'county': '$B,$row',
            'category': '$col,$6',
            'from_row': '$row',
            'from_col': '$col'}

        self.county_QNode = dict()

        # table list for every year
        self.year_table = {2006: 10, 2007: 10, 2008: 10, 2009: 10, 2010: 10, 2011: 10, 2012: 10, 2013: 10,
                           2014: 10, 2015: 10, 2016: 8, 2017: 10}

        # abbreviate?
        self.use_abbreviate = {2006: True, 2007: True, 2008: True, 2009: True, 2010: True, 2011: False, 2012: False,
                               2013: False, 2014: False, 2015: False, 2016: False, 2017: False}
        # year url format
        self.y_s_url_format = {2006: 'https://www2.fbi.gov/ucr/cius2006/data/documents/06tbl10al.xls',
                               2007: 'https://www2.fbi.gov/ucr/cius2007/data/documents/07tbl10al.xls',
                               2008: 'https://www2.fbi.gov/ucr/cius2008/data/documents/08tbl10al.xls',
                               2009: 'https://www2.fbi.gov/ucr/cius2009/data/documents/09tbl10al.xls',
                               2010: 'https://ucr.fbi.gov/crime-in-the-u.s/2010/crime-in-the-u.s.-2010/tables/table-10/10tbl10al.xls/output.xls',
                               2011: 'https://ucr.fbi.gov/crime-in-the-u.s/2011/crime-in-the-u.s.-2011/tables/table10statecuts/table_10_offenses_known_to_law_enforcement_alabama_by_metropolitan_and_nonmetropolitan_counties_2011.xls/output.xls',
                               2012: 'https://ucr.fbi.gov/crime-in-the-u.s/2012/crime-in-the-u.s.-2012/tables/10tabledatadecpdf/table-10-state-cuts/table_10_offenses_known_to_law_enforcement_alabama_by_metropolitan_and_nonmetropolitan_counties_2012.xls/output.xls',
                               2013: 'https://ucr.fbi.gov/crime-in-the-u.s/2013/crime-in-the-u.s.-2013/tables/table-10/table-10-pieces/table_10_offenses_known_to_law_enforcement_by_alabama_by_metropolitan_and_nonmetropolitan_counties_2013.xls/output.xls',
                               2014: 'https://ucr.fbi.gov/crime-in-the-u.s/2014/crime-in-the-u.s.-2014/tables/table-10/table-10-pieces/table_10_offenses_known_to_law_enforcement_alabama_by_metropolitan_and_nonmetropolitan_counties_2014.xls/output.xls',
                               2015: 'https://ucr.fbi.gov/crime-in-the-u.s/2015/crime-in-the-u.s.-2015/tables/table-10/table-10-state-pieces/table_10_offenses_known_to_law_enforcement_alabama_by_metropolitan_and_nonmetropolitan_counties_2015.xls/output.xls',
                               2016: 'https://ucr.fbi.gov/crime-in-the-u.s/2016/crime-in-the-u.s.-2016/tables/table-8/table-8-state-cuts/alabama.xls/output.xls',
                               2017: 'https://ucr.fbi.gov/crime-in-the-u.s/2017/crime-in-the-u.s.-2017/tables/table-10/table-10-state-cuts/alabama.xls/output.xls'}

        # hashmap for states and their abbrviate
        self.state_abbr = {'alabama': 'al', 'alaska': 'ak', 'arizona': 'az', 'arkansas': 'ar', 'california': 'ca',
                           'colorado': 'co', 'connecticut': 'ct', 'delaware': 'de', 'florida': 'fl', 'georgia': 'ga',
                           'hawaii': 'hi',
                           'idaho': 'id', 'illinois': 'il', 'indiana': 'in', 'iowa': 'ia', 'kansas': 'ks',
                           'kentucky': 'ky',
                           'louisiana': 'la', 'maine': 'me', 'maryland': 'md', 'massachusetts': 'ma', 'michigan': 'mi',
                           'minnesota': 'mn',
                           'mississippi': 'ms', 'missouri': 'mo', 'montana': 'mt', 'nebraska': 'ne', 'nevada': 'nv',
                           'new-hampshire': 'nh', 'new-jersey': 'nj', 'new-mexico': 'nm', 'new-york': 'ny',
                           'north-carolina': 'nc', 'north-dakota': 'nd', 'ohio': 'oh', 'oklahoma': 'ok', 'oregon': 'or',
                           'pennsylvania': 'pa',
                           'rhode-island': 'ri', 'south-carolina': 'sc', 'south-dakota': 'sd', 'tennessee': 'tn',
                           'texas': 'tx', 'utah': 'ut', 'vermont': 'vt', 'virginia': 'va', 'washington': 'wa',
                           'west-virginia': 'wv',
                           'wisconsin': 'wi', 'wyoming': 'wy'}
        # manually define kv pairs dict
        self.kv_dict = {'violent': 'C3001', 'murder': 'C3002', 'rape': 'C3003', 'robbery': 'C3005',
                        'aggravated': 'C3006', 'property': 'C3007', 'burglary': 'C3008',
                        'larceny': 'C3009', 'motor': 'C3010', 'arson': 'C3011'}

    def add_value(self, item, key, value, unit, year_value, reference):
        for property in self.kv_dict:
            if property in key:
                # add statement
                s = item.add_statement(self.kv_dict[property], QuantityValue(value, unit=unit))
                s.add_qualifier('P585', year_value)
                s.add_reference(reference)
                return
        print(key + ' is not implemented')
        raise Exception('')

    def download_data(self, years=None, states=None):

        year_list = list(self.year_table.keys())
        if years is not None:
            year_list = years

        # download all data or designated states
        state_list = list(self.state_abbr.keys())
        if states is not None:
            state_list = list()
            for s in states:
                s = s.lower().replace(' ', '-')
                state_list.append(s)
        for year in year_list:
            # delete and make new folder
            if not os.path.exists(DATA_ABS_ADDRESS + str(year)):
                os.makedirs(DATA_ABS_ADDRESS + str(year))

            for state in state_list:
                # get url
                download_url = self.y_s_url_format[int(year)]

                if self.use_abbreviate[int(year)]:
                    download_url = download_url.replace('al', self.state_abbr[state])
                else:
                    download_url = download_url.replace('alabama', state)

                # download and save excel files
                local_filename = state + '.xls'
                try:
                    with requests.get(download_url, stream=True) as r:
                        if r.status_code == 200:
                            print('Downloading crime data: ' + state + '_' + str(year))
                            with open(DATA_ABS_ADDRESS + str(year) + '/' + local_filename, 'wb') as f:

                                # save files
                                for chunk in r.iter_content():
                                    if chunk:
                                        f.write(chunk)
                except:
                    pass
        print('\n\nDownload completed!')

    def extract_data(self, years=None, states=None):

        # Initiate Excel Extractor
        ee = ExcelExtractor()
        state_list = list(self.state_abbr.keys())
        year_list = list(self.year_table.keys())
        # extract all data or designated states
        if states is not None:
            state_list = list()
            for s in states:
                s = s.lower().replace(' ', '-')
                state_list.append(s)

        # extract all data or designated years
        if years is not None:
            year_list = years
        res = dict()

        # regex to trim postfixes
        regex = 'countyunifiedpolicedepartment|,|[0-9]|policedepartment|countybureauof|countyu|publicsafety|unified|police'

        # yearwise and statewise extracting
        for year in year_list:
            for state in state_list:

                # read file
                file_path = DATA_ABS_ADDRESS + str(year) + '/' + state + '.xls'
                if not os.path.isfile(file_path):
                    continue
                else:
                    print('Extracting crime data for ' + state + '_' + str(year))

                    # extract data from excel files
                    sheet_year = '' + str(self.year_table[year]) if self.year_table[year] >= 10 else '0' + str(
                        self.year_table[year])
                    year_pre = '0' + str(year - 2000) if year - 2000 < 10 else str(year - 2000)
                    sheet_name = year_pre + 'tbl' + sheet_year + self.state_abbr[state]

                    # corner cases
                    if year < 2008: sheet_name = 'Sheet1'
                    if year == 2011 and state == 'oregon': sheet_name = '11tbl10'
                    extractions = ee.extract(file_name=file_path,
                                             sheet_name=sheet_name,
                                             region=['B,7', 'N,100'],
                                             variables=self.value_dict)
                    # build dictionary
                    county_data = dict()
                    for e in extractions:
                        if len(e['county']) > 0:

                            # clean data
                            name = e['county'].replace(' ', '').lower()
                            if '/' in name:
                                name = name.split('/')[1]
                            if '-' in name:
                                name = name.split('-')[1]

                            # trim bunch of postfixes
                            name = re.sub(regex, '', name)

                            # build county dictionary
                            if name not in county_data:
                                county_data[name] = dict()

                            # add extracted data
                            if e['value'] is not '' and isinstance(e['value'], int):

                                # clean data
                                ctg = e['category'].replace('\n', '_').replace(' ', '_').lower()
                                if ctg[-1].isdigit():
                                    ctg = ctg[:-1]
                                if ctg in county_data[name]:
                                    county_data[name][ctg] += e['value']
                                else:
                                    county_data[name][ctg] = e['value']
                    res[state + '_' + str(year)] = county_data
        print('\n\nExtraction completed!')
        return res

    def model_data(self, crime_data, file_path, format='ttl'):

        # initialize
        kg_schema = KGSchema()
        kg_schema.add_schema('@prefix : <http://isi.edu/> .', 'ttl')
        etk = ETK(kg_schema=kg_schema, modules=ETKModule)
        doc = etk.create_document({}, doc_id="http://isi.edu/default-ns/projects")

        # bind prefixes
        doc.kg.bind('wikibase', 'http://wikiba.se/ontology#')
        doc.kg.bind('wd', 'http://www.wikidata.org/entity/')
        doc.kg.bind('wdt', 'http://www.wikidata.org/prop/direct/')
        doc.kg.bind('wdtn', 'http://www.wikidata.org/prop/direct-normalized/')
        doc.kg.bind('wdno', 'http://www.wikidata.org/prop/novalue/')
        doc.kg.bind('wds', 'http://www.wikidata.org/entity/statement/')
        doc.kg.bind('wdv', 'http://www.wikidata.org/value/')
        doc.kg.bind('wdref', 'http://www.wikidata.org/reference/')
        doc.kg.bind('p', 'http://www.wikidata.org/prop/')
        doc.kg.bind('pr', 'http://www.wikidata.org/prop/reference/')
        doc.kg.bind('prv', 'http://www.wikidata.org/prop/reference/value/')
        doc.kg.bind('prn', 'http://www.wikidata.org/prop/reference/value-normalized/')
        doc.kg.bind('ps', 'http://www.wikidata.org/prop/statement/')
        doc.kg.bind('psv', 'http://www.wikidata.org/prop/statement/value/')
        doc.kg.bind('psn', 'http://www.wikidata.org/prop/statement/value-normalized/')
        doc.kg.bind('pq', 'http://www.wikidata.org/prop/qualifier/')
        doc.kg.bind('pqv', 'http://www.wikidata.org/prop/qualifier/value/')
        doc.kg.bind('pqn', 'http://www.wikidata.org/prop/qualifier/value-normalized/')
        doc.kg.bind('skos', 'http://www.w3.org/2004/02/skos/core#')
        doc.kg.bind('prov', 'http://www.w3.org/ns/prov#')
        doc.kg.bind('schema', 'http://schema.org/')

        # first we add properties and entities
        # Define Qnode for properties related to crime.
        q = WDItem('D1001')
        q.add_label('Wikidata property related to crime', lang='en')
        q.add_statement('P279', Item('Q22984475'))
        q.add_statement('P1269', Item('Q83267'))
        doc.kg.add_subject(q)

        # violent crime offenses
        p = WDProperty('C3001', Datatype.QuantityValue)
        p.add_label('violent crime offenses', lang='en')
        p.add_description(
            "number of violent crime offenses reported by the sheriff's office or county police department", lang='en')
        p.add_statement('P31', Item('D1001'))
        p.add_statement('P1629', Item('Q1520311'))
        doc.kg.add_subject(p)

        # murder and non - negligent manslaughter
        p = WDProperty('C3002', Datatype.QuantityValue)
        p.add_label('murder and non-negligent manslaughter', lang='en')
        p.add_description(
            "number of murder and non-negligent manslaughter offenses reported by the sheriff's office or county police department",
            lang='en')
        p.add_statement('P31', Item('D1001'))
        p.add_statement('P1629', Item('Q1295558'))
        p.add_statement('P1629', Item('Q132821'))
        doc.kg.add_subject(p)

        # Rape(revised definition)
        p = WDProperty('C3003', Datatype.QuantityValue)
        p.add_label('Rape (revised definition)', lang='en')
        p.add_description(
            "number of rapes (revised definition) reported by the sheriff's office or county police department",
            lang='en')
        p.add_statement('P31', Item('D1001'))
        p.add_statement('P1629', Item('Q47092'))
        doc.kg.add_subject(p)

        # robbery
        p = WDProperty('C3005', Datatype.QuantityValue)
        p.add_label('Robbery', lang='en')
        p.add_description("number of roberies reported by the sheriff's office or county police department", lang='en')
        p.add_statement('P31', Item('D1001'))
        p.add_statement('P1629', Item('Q53706'))
        doc.kg.add_subject(p)

        # aggravated assault
        p = WDProperty('C3006', Datatype.QuantityValue)
        p.add_label('Aggravated assault', lang='en')
        p.add_description("number of aggravated assaults reported by the sheriff's office or county police department",
                          lang='en')
        p.add_statement('P31', Item('D1001'))
        p.add_statement('P1629', Item('Q365680'))
        p.add_statement('P1629', Item('Q81672'))
        doc.kg.add_subject(p)

        # property crime
        p = WDProperty('C3007', Datatype.QuantityValue)
        p.add_label('Property crime', lang='en')
        p.add_description("number of property crimes reported by the sheriff's office or county police department",
                          lang='en')
        p.add_statement('P31', Item('D1001'))
        p.add_statement('P1629', Item('Q857984'))
        doc.kg.add_subject(p)

        # burglary
        p = WDProperty('C3008', Datatype.QuantityValue)
        p.add_label('Burglary', lang='en')
        p.add_description("number of Burglaries reported by the sheriff's office or county police department",
                          lang='en')
        p.add_statement('P31', Item('D1001'))
        p.add_statement('P1629', Item('Q329425'))
        doc.kg.add_subject(p)

        # larceny - theft
        p = WDProperty('C3009', Datatype.QuantityValue)
        p.add_label('Larceny-theft', lang='en')
        p.add_description("number of Larceny-theft reported by the sheriff's office or county police department",
                          lang='en')
        p.add_statement('P31', Item('D1001'))
        p.add_statement('P1629', Item('Q2485381'))
        p.add_statement('P1629', Item('Q2727213'))
        doc.kg.add_subject(p)

        # motor vehicle theft
        p = WDProperty('C3010', Datatype.QuantityValue)
        p.add_label('Motor vehicle theft', lang='en')
        p.add_description("number of Motor vehicle thefts reported by the sheriff's office or county police department",
                          lang='en')
        p.add_statement('P31', Item('D1001'))
        p.add_statement('P1629', Item('Q548007'))
        p.add_statement('P1629', Item('Q2727213'))
        doc.kg.add_subject(p)

        # arson
        p = WDProperty('C3011', Datatype.QuantityValue)
        p.add_label('Arson', lang='en')
        p.add_description("number of arsons reported by the sheriff's office or county police department", lang='en')
        p.add_statement('P31', Item('D1001'))
        p.add_statement('P1629', Item('Q327541'))
        doc.kg.add_subject(p)

        # Offenses are reported for a period of type,
        # so the quantity needs to be represented in units such as offenses / year
        unit = WDItem('D1002')
        unit.add_label('offenses per year', lang='en')
        unit.add_statement('P31', Item('Q47574'))
        unit.add_statement('P1629', Item('Q83267'))
        # doc.kg.add_subject(unit)

        # we begin to model data extracted
        for state_year in crime_data:
            print('Modeling data for ' + state_year)
            state, year = state_year.split('_')

            # add year value
            year_value = TimeValue(year, calendar=Item('Q1985727'), precision=Precision.year, time_zone=0)

            # add reference, data source
            download_url = 'https://ucr.fbi.gov/crime-in-the-u.s/' + str(year) + '/crime-in-the-u.s.-' + str(
                year) + '/tables/table-' + str(self.year_table[int(year)]) + '/table-' + str(
                self.year_table[int(year)]) + '-state-cuts/' + str(state) + '.xls'
            reference = WDReference()
            reference.add_property(URI('P248'), URI('wd:Q8333'))
            reference.add_property(URI('P854'), Literal(download_url))

            for county in crime_data[state_year]:

                # county entity
                Qnode = self.get_QNode(county, state)
                if Qnode is None:
                    continue
                q = WDItem(Qnode)

                # add value for each property
                for property in crime_data[state_year][county]:
                    self.add_value(q, property, crime_data[state_year][county][property], unit, year_value,
                                   reference)

                # add the entity to kg
                doc.kg.add_subject(q)
        print('\n\nModeling completed!\n\n')

        # serialization
        f = open(file_path, 'w')
        f.write(doc.kg.serialize(format))
        f.close()

        print('Serialization completed!')

    def query_QNodes(self):

        # query QNode for each county
        endpoint_url = "https://query.wikidata.org/sparql"

        query = """SELECT ?QNode ?name ?state
                    WHERE { ?QNode  wdt:P882 ?v .
                            ?QNode rdfs:label ?name.
                            ?QNode wdt:P131 ?sq.
                            ?sq rdfs:label ?state.
                          FILTER(LANG(?name) = "en")
                          FILTER(LANG(?state) = "en")}
                """

        # get Qnodes from wikidata
        def get_results(endpoint_url, query):
            sparql = SPARQLWrapper(endpoint_url)
            sparql.setQuery(query)
            sparql.setReturnFormat(JSON)
            return sparql.query().convert()

        results = get_results(endpoint_url, query)
        county_Q = dict()

        # concatenate county name and state and build dict
        for result in results["results"]["bindings"]:
            name = result['name']['value'].lower().replace(' ', '')
            if 'parish' in name:
                name = name[:-6]
            state = result['state']['value'].lower().replace(' ', '-')
            if state in self.state_abbr:
                abbr = self.state_abbr[state]
                QNode = result['QNode']['value'].split('/')[-1]
                county_Q[name + '_' + abbr] = QNode

        # save json file
        f = open('county_QNode.json', 'w')
        f.write(json.dumps(county_Q, indent=4))
        f.close()
        print('\n\nCounty - QNode mapping dictionary created\n\n')
        return county_Q

    def get_QNode(self, county, state):
        # read county_QNode.json to build dictionary
        if len(self.county_QNode) == 0:
            f = open('county_QNode.json', 'r')
            self.county_QNode = json.loads(f.read())
            f.close()
        name = county.lower().replace(' ', '')

        temp_state = state.lower().replace(' ', '-')
        # map state to its abbreviate
        if temp_state in self.state_abbr:
            abbr = self.state_abbr[temp_state]

            # find Qnode
            county1 = name + 'county_' + abbr
            if county1 in self.county_QNode:
                return self.county_QNode[county1]

            # corner cases
            if county1 == 'dadecounty_fl':
                return 'Q468557'
            if county1 == 'donaanacounty_nm':
                return 'Q112953'
            if county1 == 'oglalalakotacounty_sd':
                return 'Q495201'
            if county1 == 'bartholemewcounty_in':
                return 'Q504385'
            if county1 == 'pottawattamiecounty_ks':
                return 'Q376947'
            if name == 'st.helena' and temp_state == 'louisiana':
                return 'Q507112'
            if name == 'story' and temp_state == 'nevada':
                return 'Q484431'
            if name == 'allegheny' and temp_state == 'maryland':
                return 'Q156257'
            if name == 'allegany' and temp_state == 'michigan':
                return 'Q133860'

            county2 = name + '_' + abbr
            if county2 in self.county_QNode:
                return self.county_QNode[county2]

        return None


if __name__ == "__main__":
    import warnings
    warnings.filterwarnings("ignore")
    parser = argparse.ArgumentParser(description='Year list')
    parser.add_argument('option', type=int,
                        help='1 for download, 2 for extraction and modeling', default=1)

    parser.add_argument('--years', type=str,
                        help='years you want to extract (default: all)', default="all")

    parser.add_argument('--states', type=str, default='all',
                        help='states you want to extract (default: "all")')

    args = parser.parse_args()

    choice = args.option
    years_str = args.years

    if years_str == 'all':
        years = None
    else:
        years = [int(x.strip()) for x in years_str.split(',')]
    states_str = args.states
    if states_str == 'all':
        states = None
    else:
        states = [x.strip() for x in states_str.split(',')]
    model = FBI_Crime_Model()

    # run once to get QNodes dictionary
    model.query_QNodes()

    if choice == 1:
        model.download_data(years, states)
    elif choice == 2:
        res = model.extract_data(years, states)
        model.model_data(res, 'result.ttl')
        with open('changes.tsv', 'w') as fp:
            serialize_change_record(fp)
    else:
        print("Option: 1 for download, 2 for extraction and modeling")



from datamart_isi.profilers.basic_profiler import BasicProfiler
from datamart_isi.profilers.dsbox_profiler import DSboxProfiler
from datamart_isi.profilers.two_ravens_profiler import TwoRavensProfiler


class Profiler(object):

    def __init__(self):
        self.basic_profiler = BasicProfiler()
        self.dsbox_profiler = DSboxProfiler()
        self.two_ravens_profiler = TwoRavensProfiler()



from datasketch import MinHash
from collections import defaultdict
import os, pickle


# input IDs, return MinHash Object
def minhash_IDs(IDs):
    m = MinHash()
    for ID in IDs:
        m.update(ID.encode('utf-8'))
    return m


if __name__ == "__main__":
    # iterate through this array, do minhash
    filenames = os.listdir("./results_2.0")
    minhash_list = []
    rank_list = []
    for idx,filename in enumerate(filenames):
        cur_file = open('./results_2.0/' + filename,'r')
        try:
            cur_ids = cur_file.read().split('\n')
            cur_file.close()
            minhash_list.append((minhash_IDs(cur_ids),filename))
        except:
            print(filename)
        cur_file.close()
        # if(idx>300):
        #     break

    # provide a filename, then compare jaccard similarity with all of the files
    #filename = 'P5942_Protected objects Ostbelgien ID.txt'
    #cur_file = open('./results_2.0/' + filename,'r')
    #cur_file = open('/Users/tong/Desktop/missing_results/P846_Global Biodiversity Information Facility ID.txt','r')
    cur_file = open('./test.txt','r')
    
    cur_ids = cur_file.read().split('\n')
    cur_m = minhash_IDs(cur_ids)
    cur_file.close()

    cur_file = open('./test.txt','r')
    cur_id = cur_file.read().split('\n')
    cur_n = minhash_IDs(cur_id)
    print(cur_m.jaccard(cur_n))
    for mHash, name in minhash_list:
        similarity = cur_m.jaccard(mHash)
        rank_list.append((similarity, name))
    rank_list.sort(reverse = True)
    for sim, fname in rank_list[:5]:
        print('similarity : '+str(sim)+' name: '+fname)



''' references

https://ekzhu.github.io/datasketch/minhash.html
https://blog.csdn.net/lsq2902101015/article/details/51305825
https://python3-cookbook.readthedocs.io/zh_CN/latest/c05/p21_serializing_python_objects.html
https://blog.csdn.net/Lycoridiata/article/details/78536768
'''
'''store_minhash(minhash_list,'minhash_list.pkl')
f = open('minhash_list.pkl', 'rb')
minhash_list2 = pickle.loads(f)
print(minhash_list2)
f.close()'''
# store minhash list into pickle
# def store_minhash(data, file_name):
#     f = open(file_name, "wb")
#     pickle.dump(data, f)



# import os
# from .sparql import DEFAULT_SAVE_LOCATION, save_prop_idents
from .find_identity import FindIdentity
import typing
import numpy as np
from collections import Counter
import copy
import math
# class Wikifier:
#     def __init__(self, base_file_loc:str=None):
#         if base_file_loc:
#             pass
#         else:
#             print("No base file given, try to find the base file on default location")
#             if os.path.exists(DEFAULT_SAVE_LOCATION):
#                 pass
#                 # with open(DEFAULT_SAVE_LOCATION, "r")
#             else:
#                 print(" No base file found on default save location please run Wikifier.initialize() first")
#
#     @staticmethod
#     def initialize():
#         save_prop_idents()


def produce(inputs, target_columns: typing.List[int]=None, target_p_node: typing.List[str]=None, input_type: str="pandas"):
    if input_type == "pandas":
        return produce_for_pandas(input_df=inputs, target_columns=target_columns, target_p_node=target_p_node)
    # elif input_type == "d3m_ds":
    #     return produce_for_d3m_dataset(input_ds=inputs, target_columns=target_columns)
    # elif input_type == "d3m_df":
    #     return produce_for_d3m_dataframe(input_df=inputs, target_columns=target_columns)
    else:
        raise ValueError("unknown type of input!")

def all_in_range_0_to_100(inputs):
    min_val = min(inputs)
    max_val = max(inputs)
    if min_val <= 100 and min_val >= 0 and max_val >= 0 and max_val <= 100:
        return True
    else:
        return False

def are_almost_continues_numbers(inputs, threshold=0.7):
    min_val = min(inputs)
    max_val = max(inputs)
    if (max_val - min_val) * threshold <= len(inputs):
        return True
    else:
        return False

def produce_for_pandas(input_df, target_columns: typing.List[int]=None, target_p_node: typing.List[str]=None, threshold_for_converage=0.7):
    """
    function used to produce for input type is pandas.dataFrame
    :param input_df: input pd.dataFrame
    :param target_columns: target columns to find with wikidata
    :param target_p_node: user-speicified P node want to get, can be None if want automatic search
    :param threshold_for_converage: minimum coverage ratio for a wikidata columns to be appended
    :return: a pd.dataFrame with updated columns from wikidata
    """
    # if no target columns given, just try every str columns
    if target_columns is None:
        target_columns = list(range(input_df.shape[1]))

    return_df = copy.deepcopy(input_df)
    for column in target_columns:
        current_column_name = input_df.columns[column]
        # curData = [str(x) for x in list(input_df[column])]
        print('Current column: ' + current_column_name)
        try:
            temp = set()
            for each in input_df[current_column_name].dropna().tolist():
                temp.add(int(each))
            if all_in_range_0_to_100(temp) or are_almost_continues_numbers(temp, threshold_for_converage):
                print("A columns with all numerical values and useless detected, skipped")
                continue
        except:
            pass

        curData = []
        for each in input_df.iloc[:, column].tolist():
            if type(each) is str:
                curData.append(each)
            elif each is np.nan or math.isnan(each):
                curData.append("")
            else:
                curData.append(str(each))

        if coverage(curData) < threshold_for_converage:
            print("[WARNING] Coverage of data is " + str(coverage(curData)) + " which is less than threshold " + str(threshold_for_converage))
            continue
        # for each column, try to find corresponding possible P nodes id first
        if target_p_node is not None:
            target_p_node_to_send = target_p_node[column]
        else:
            target_p_node_to_send = None
        for idx, res in enumerate(FindIdentity.get_identifier_3(strings=curData, column_name=current_column_name,
                                                                target_p_node=target_p_node_to_send)):
            # res[0] is the send input P node
            top1_dict = res[1]
            new_col = [""] * len(curData)
            for i in range(len(curData)):
                if curData[i] in top1_dict:
                    new_col[i] = top1_dict[curData[i]]
            if coverage(new_col) < threshold_for_converage:
                print(
                    "[WARNING] Coverage of Q nodes is " + str(coverage(new_col)) + " which is less than threshold " + str(threshold_for_converage))
                continue
            col_name = current_column_name + '_wikidata'
            return_df[col_name] = new_col
            break
    return return_df


def coverage(column):
    count_stats = Counter(column)
    return (len(column)-count_stats[''])/len(column)



from abc import ABC, abstractmethod
import pandas as pd
from enum import Enum
import typing
from datamart_isi.joiners.join_result import JoinResult


class JoinerBase(ABC):
    """Abstract class of Joiner, should be extended for other joiners.

    """

    @abstractmethod
    def join(self, **kwargs) -> JoinResult:
        """Implement join method which returns a pandas Dataframe

        """
        pass


class JoinerType(Enum):
    DEFAULT = "default"
    RLTK = "rltk"
    EXACT_MATCH = "exact_match"


class DefaultJoiner(JoinerBase):
    """
    Default join class.
    Do exact join by using pandas merge join function
    """

    @staticmethod
    def join(left_df: pd.DataFrame,
             right_df: pd.DataFrame,
             left_columns: typing.List[typing.List[int]],
             right_columns: typing.List[typing.List[int]],
             **kwargs
             ) -> JoinResult:

        left_columns = [x[0] for x in left_columns]
        right_columns = [x[0] for x in right_columns]

        if len(left_columns) != len(right_columns):
            raise ValueError("Default join only perform on 1-1 mapping")

        right_df = right_df.rename(columns={
            right_df.columns[right_columns[idx]]: left_df.columns[left_columns[idx]] for idx in range(len(left_columns))
        })

        df = pd.merge(left=left_df,
                        right=right_df,
                        left_on=[left_df.columns[idx] for idx in left_columns],
                        right_on=[right_df.columns[idx] for idx in right_columns],
                        how='left')

        return JoinResult(df=df)


class JoinerPrepare(object):

    @staticmethod
    def prepare_joiner(joiner: JoinerType = JoinerType.DEFAULT) -> typing.Optional[JoinerBase]:

        """Prepare joiner, lazy evaluation for joiners,
        should be useful because joiner like RLTK may need many dependency packages.

        Args:
            joiner: string of joiner type

        Returns:
            joiner instance

        """

        try:
            JoinerType(joiner)
        except ValueError:
            return None

        if JoinerType(joiner) == JoinerType.RLTK:
            from datamart_isi.joiners.rltk_joiner import RLTKJoiner
            return RLTKJoiner()

        if JoinerType(joiner) == JoinerType.DEFAULT:
            return DefaultJoiner()

        if JoinerType(joiner) == JoinerType.EXACT_MATCH:
            from datamart_isi.joiners.exact_match_joiner import ExactMatchJoiner
            return ExactMatchJoiner()

        return None



import pandas as pd
# import os
import copy
import random
import frozendict
import collections
import typing
from d3m.container import DataFrame as d3m_DataFrame
from d3m.container import Dataset as d3m_Dataset
import d3m.metadata.base as metadata_base
# from datamart.dataset import Dataset
from datamart.utilities.utils import PRODUCTION_ES_INDEX, SEARCH_URL
from datamart.es_managers.json_query_manager import JSONQueryManager
from datamart.new_query.augment import Augment
# old Augment 
# from datamart.augment import Augment
# from datamart.data_loader import DataLoader
from d3m.base import utils as d3m_utils
from datamart.utilities.utils import Utils
# from datamart.joiners.join_result import JoinResult
# from datamart.joiners.joiner_base import JoinerType
# from itertools import chain
from datamart.joiners.rltk_joiner import RLTKJoiner
from SPARQLWrapper import SPARQLWrapper, JSON, POST, URLENCODED
from d3m.metadata.base import DataMetadata, ALL_ELEMENTS
from datamart.joiners.rltk_joiner import RLTKJoiner_new
from wikifier import config
# import requests
import traceback
import logging
import datetime
import enum

Q_NODE_SEMANTIC_TYPE = "http://wikidata.org/qnode"
DEFAULT_URL = "https://isi-datamart.edu"
__ALL__ = ('D3MDatamart', 'DatamartSearchResult', 'D3MJoinSpec')
DatamartSearchResult = typing.TypeVar('DatamartSearchResult', bound='DatamartSearchResult')
D3MJoinSpec = typing.TypeVar('D3MJoinSpec', bound='D3MJoinSpec')
DatamartQuery = typing.TypeVar('DatamartQuery', bound='DatamartQuery')
MAX_ENTITIES_LENGTH = 200
CONTAINER_SCHEMA_VERSION = 'https://metadata.datadrivendiscovery.org/schemas/v0/container.json'
P_NODE_IGNORE_LIST = {"P1549"}
SPECIAL_REQUEST_FOR_P_NODE = {"P1813": "FILTER(strlen(str(?P1813)) = 2)"}
AUGMENT_RESOURCE_ID = "augmentData"


class D3MDatamart:
    """
    ISI implementation of D3MDatamart
    """
    def __init__(self, mode):
        self.url = DEFAULT_URL
        self._logger = logging.getLogger(__name__)
        if mode == "test":
            query_server = config.endpoint_query_test
        else:
            query_server = config.endpoint_query_main
        self.augmenter = Augment(endpoint=query_server)

    def search(self, query: DatamartQuery, timeout=None, limit: int = 20) -> typing.List[DatamartSearchResult]:
        """
        This entry point supports search using a query specification.
        The query spec is rich enough to enable query by example. The caller can select
        values from a column in a dataset and use the query spec to find datasets that can join with the values
        provided. Use of the query spec enables callers to compose their own "smart search" implementations.

        Parameters
        ----------
        query: DatamartQuery
            Query specification.
        timeout: int
            Maximum number of seconds before returning results.
        limit: int
            Maximum number of search results to return.

        returns
        -------
        typing.List[DatamartSearchResult]
            List of search results, combined from multiple datamarts. In the baseline implementation the list of
            datamarts will be randomized on each call, and the results from multiple datamarts will be interleaved
            accordingly. There will be an attempt to deduplicate results from multiple datamarts.
        """
        if not self.url.startswith(SEARCH_URL):
            return []
        query_json = query.to_json()
        return self.search_general(query=query_json, supplied_data=None, timeout=timeout, limit=limit)

    def search_general(self, query, supplied_data: d3m_DataFrame=None, timeout=None, limit: int=20) \
            -> typing.List[DatamartSearchResult]:
        """
        The search function used for general elastic search
        :param query: JSON object describing the query.
        :param supplied_data: the data you are trying to augment.
        :param timeout: allowed time spent on searching
        :param limit: the limitation on the return amount of DatamartSearchResult
        :return: list of search results of DatamartSearchResult
        """
        query_results = []
        try:
            params = {"timeout": "2m"}
            if (query and ('required_variables' in query)) or (supplied_data is None):
                # if ("required_variables" exists or no data):

                cur_results = self.augmenter.query_by_sparql(query, supplied_data, size=limit) or []
                for res in cur_results:
                    query_results.append(DatamartSearchResult(search_result=res, supplied_data=supplied_data,
                                                              query_json=query, search_type="general")
                                         )

            else:
                # if there is no "required_variables" in the query JSON, but the supplied_data exists,
                # try each named entity column as "required_variables" and concat the results:
                query = query or {}
                exist = set()
                for col in supplied_data:
                    if Utils.is_column_able_to_query(supplied_data[col]):
                        # update 2019.4.9: we should not replace the original query!!!
                        query_copy = query.copy()
                        query_copy['required_variables'] = [{
                            "type": "dataframe_columns",
                            "names": [col]
                        }]
                        cur_results = self.augmenter.query_by_sparql(query_copy, supplied_data, size=limit, params=params)
                        if not cur_results:
                            continue
                        for res in cur_results:
                            if res['_id'] not in exist:
                                # TODO: how about the score ??
                                exist.add(res['_id'])
                            query_results.append(DatamartSearchResult(search_result=res, supplied_data=supplied_data,
                                                                   query_json=query_copy, search_type="general")
                                              )
            return query_results
        except:
            print("Searching with wiki data failed")
            traceback.print_exc()
        finally:
            return query_results

    def search_with_data(self, supplied_data: typing.Union[d3m_Dataset, d3m_DataFrame], query: DatamartQuery = None,
                         timeout=None, limit: int = 20) -> typing.List[DatamartSearchResult]:
        """
        This entry point supports search based on a supplied datasets.

        *) Smart search: the caller provides supplied_data (a D3M dataset), and a query containing
        keywords from the D3M problem specification. The search will analyze the supplied data and look for datasets
        that can augment the supplied data in some way. The keywords are used for further filtering and ranking.
        For example, a datamart may try to identify named entities in the supplied data and search for companion
        datasets that contain these entities.

        *) Columns search: this search is similar to smart search in that it uses both query spec and the supplied_data.
        The difference is that only the data in the columns listed in query_by_example_columns is used to identify
        companion datasets.

        Parameters
        ---------
        query: DatamartQuery
            Query specification
        supplied_data: d3m.container.Dataset or d3m.container.DataFrame
            The data you are trying to augment.
        query_by_example_columns:
            A list of column identifiers or column names.
        timeout: int
            Maximum number of seconds before returning results.
        limit: int
            Maximum number of search results to return.


        Returns
        -------
        typing.List[DatamartSearchResult]
            A list of DatamartSearchResult of possible companion datasets for the supplied data.
        """

        if not self.url.startswith(SEARCH_URL):
            return []

        if type(supplied_data) is d3m_Dataset:
            res_id, supplied_dataframe = d3m_utils.get_tabular_resource(dataset=supplied_data, resource_id=None)
        else:
            supplied_dataframe = supplied_data

        search_results = []
        if query is not None:
            query_json = query.to_json()
        else:
            query_json = None

        # first take a search on wikidata
        wikidata_search_results = self.search_wiki_data(query_json, supplied_data, timeout)
        search_results.extend(wikidata_search_results)
        limit_remained = limit - len(wikidata_search_results)

        if query is None:
            # if not query given, try to find the Text columns from given dataframe and use it to find some candidates
            can_query_columns = []
            for each in range(len(supplied_dataframe.columns)):
                if type(supplied_data) is d3m_Dataset:
                    selector = (res_id, ALL_ELEMENTS, each)
                else:
                    selector = (ALL_ELEMENTS, each)
                each_column_meta = supplied_data.metadata.query(selector)
                if 'http://schema.org/Text' in each_column_meta["semantic_types"]:
                    # or "https://metadata.datadrivendiscovery.org/types/CategoricalData" in each_column_meta["semantic_types"]:
                    can_query_columns.append(each)

            # import pdb
            # pdb.set_trace()

            if len(can_query_columns) == 0:
                self._logger.warning("No columns can be augment!")
                return search_results

            results_no_query = []
            for each_column in can_query_columns:
                column_values = supplied_dataframe.iloc[:, each_column]
                query_column_entities = list(set(column_values.tolist()))

                if len(query_column_entities) > MAX_ENTITIES_LENGTH:
                    query_column_entities = random.sample(query_column_entities, MAX_ENTITIES_LENGTH)

                for i in range(len(query_column_entities)):
                    query_column_entities[i] = str(query_column_entities[i])

                query_column_entities = " ".join(query_column_entities)

                search_query = DatamartQuery(about=query_column_entities)
                query_json = search_query.to_json()
                # TODO: need to improve the query for required variables
                # search_query ={
                #     "required_variables": [
                #         {
                #             "type": "generic_entity",
                #             "named_entities":query_column_entities
                #         }
                #     ]
                # }

                # sort to put best results at top
                temp_results = self.search_general(query=query_json, supplied_data=supplied_dataframe)
                temp_results.sort(key=lambda x: x.score, reverse=True)
                results_no_query.append(temp_results)
            # we will return the results of each searching query one by one
            # for example: [res_q1_1,res_q1_2,res_q1_3], [res_q2_1,res_q2_2,res_q2_3] , [res_q3_1,res_q3_2,res_q3_3]
            # will return as: [res_q1_1, res_q2_1, res_q3_1, res_q1_2, res_q2_2, res_q3_3...]
            results_rescheduled = []
            has_remained = True
            while has_remained:
                has_remained = False
                for each in results_no_query:
                    if len(each) > 0:
                        has_remained = True
                        results_rescheduled.append(each.pop(0))
            # append together
            search_results.extend(results_rescheduled)

        else:  # for the condition if given query, follow the query
            if limit_remained > 0:
                query_json = query.to_json()
                general_search_results = self.search_general(query=query_json, supplied_data=supplied_dataframe,
                                                             timeout=timeout, limit=limit_remained)
                general_search_results.sort(key=lambda x: x.score, reverse=True)
                search_results.extend(general_search_results)

        if len(search_results) > limit:
            search_results = search_results[:limit]
        return search_results

    def search_with_data_columns(self, supplied_data: container.Dataset, data_constraints: typing.List['TabularVariable'], 
                                 query: 'DatamartQuery'=None) -> "DatamartQueryCursor":
        """
        Search using a query which can include constraints on supplied data columns (TabularVariable).

        This search is similar to the "smart" search provided by `search_with_data()`, but caller must manually specify
        constraints using columns from the supplied data; Datamart will not automatically analyze it to determine
        relevance or joinability.

        Use of the query spec enables callers to compose their own "smart search" implementations.

        Datamart implementations should return a DatamartQueryCursor immediately.

        Parameters
        ------_---
        query : DatamartQuery
            Query specification
        supplied_data : container.Dataset
            The data you are trying to augment.
        data_constraints : list
            List of `TabularVariable` constraints referencing the supplied data.

        Returns
        -------
        DatamartQueryCursor
            A cursor pointing to search results containing possible companion datasets for the supplied data.
        """
        if query is None:
            


    def augment(self, query, supplied_data: d3m_Dataset, timeout: int = None, max_new_columns: int = 1000) -> d3m_Dataset:
        """
        In this entry point, the caller supplies a query and a dataset, and datamart returns an augmented dataset.
        Datamart automatically determines useful data for augmentation and automatically joins the new data to produce
        an augmented Dataset that may contain new columns and rows, and possibly new dataframes.

        Parameters
        ---------
        query: DatamartQuery
            Query specification
        supplied_data: d3m.container.Dataset
            The data you are trying to augment.
        timeout: int
            Maximum number of seconds before returning results.
        max_new_columns: int
            Maximum number of new columns to add to the original dataset.

        Returns
        -------
        d3m.container.Dataset
            The augmented Dataset
        """
        if type(supplied_data) is d3m_Dataset:
            input_type = "ds"
            res_id, _ = d3m_utils.get_tabular_resource(dataset=supplied_data, resource_id=None)
        else:
            input_type = "df"

        search_results = self.search_with_data(query=query, supplied_data=supplied_data, timeout=timeout)
        continue_aug = True
        count = 0
        augment_result = supplied_data

        # continue augmenting until reach the maximum new columns or all search_result has been used
        while continue_aug and count < len(search_results):
            augment_result = search_results[count].augment(supplied_data=augment_result)
            count += 1
            if input_type == "ds":
                current_column_number = augment_result[res_id].shape[1]
            else:
                current_column_number = augment_result.shape[1]
            if current_column_number >= max_new_columns:
                continue_aug = False

        return augment_result

    @staticmethod
    def search_wiki_data(query, supplied_data: typing.Union[d3m_DataFrame, d3m_Dataset]=None, timeout=None,
                         search_threshold=0.5) -> typing.List[DatamartSearchResult]:
        """
        The search function used for wikidata search
        :param query: JSON object describing the query.
        :param supplied_data: the data you are trying to augment.
        :param timeout: allowed time spent on searching
        :param limit: the limitation on the return amount of DatamartSearchResult
        :param search_threshold: the minimum appeared times of the properties
        :return: list of search results of DatamartSearchResult
        """
        wikidata_results = []
        try:
            q_nodes_columns = []
            if type(supplied_data) is d3m_Dataset:
                res_id, supplied_dataframe = d3m_utils.get_tabular_resource(dataset=supplied_data, resource_id=None)
                selector_base_type = "ds"
            else:
                supplied_dataframe = supplied_data
                selector_base_type = "df"

            # check whether Qnode is given in the inputs, if given, use this to wikidata and search
            required_variables_names = None
            metadata_input = supplied_data.metadata

            if query is not None and 'required_variables' in query:
                required_variables_names = []
                for each in query['required_variables']:
                    required_variables_names.extend(each['names'])
            for i in range(supplied_dataframe.shape[1]):
                if selector_base_type == "ds":
                    metadata_selector = (res_id, metadata_base.ALL_ELEMENTS, i)
                else:
                    metadata_selector = (metadata_base.ALL_ELEMENTS, i)
                if Q_NODE_SEMANTIC_TYPE in metadata_input.query(metadata_selector)["semantic_types"]:
                    # if no required variables given, attach any Q nodes found
                    if required_variables_names is None:
                        q_nodes_columns.append(i)
                    # otherwise this column has to be inside required_variables
                    else:
                        if supplied_dataframe.columns[i] in required_variables_names:
                            q_nodes_columns.append(i)

            if len(q_nodes_columns) == 0:
                print("No wikidata Q nodes detected on corresponding required_variables! Will skip wikidata search part")
                return wikidata_results
            else:

                print("Wikidata Q nodes inputs detected! Will search with it.")
                print("Totally " + str(len(q_nodes_columns)) + " Q nodes columns detected!")

                # do a wikidata search for each Q nodes column
                for each_column in q_nodes_columns:
                    q_nodes_list = supplied_dataframe.iloc[:, each_column].tolist()
                    p_count = collections.defaultdict(int)
                    p_nodes_needed = []
                    # temporary block
                    """
                    http_address = 'http://minds03.isi.edu:4444/get_properties'
                    headers = {"Content-Type": "application/json"}
                    requests_data = str(q_nodes_list)
                    requests_data = requests_data.replace("'", '"')
                    r = requests.post(http_address, data=requests_data, headers=headers)
                    results = r.json()
                    for each_p_list in results.values():
                        for each_p in each_p_list:
                            p_count[each_p] += 1
                    """
                    # TODO: temporary change here, may change back in the future
                    # Q node format (wd:Q23)(wd: Q42)
                    q_node_query_part = ""
                    unique_qnodes = set(q_nodes_list)
                    for each in unique_qnodes:
                        if len(each) > 0:
                            q_node_query_part += "(wd:" + each + ")"
                    sparql_query = "select distinct ?item ?property where \n{\n  VALUES (?item) {" + q_node_query_part \
                                   + "  }\n  ?item ?property ?value .\n  ?wd_property wikibase:directClaim ?property ." \
                                   + "  values ( ?type ) \n  {\n    ( wikibase:Quantity )\n" \
                                   + "    ( wikibase:Time )\n    ( wikibase:Monolingualtext )\n  }" \
                                   + "  ?wd_property wikibase:propertyType ?type .\n}\norder by ?item ?property "

                    try:
                        sparql = SPARQLWrapper(WIKIDATA_QUERY_SERVER)
                        sparql.setQuery(sparql_query)
                        sparql.setReturnFormat(JSON)
                        sparql.setMethod(POST)
                        sparql.setRequestMethod(URLENCODED)
                        results = sparql.query().convert()['results']['bindings']
                    except:
                        print("Getting query of wiki data failed!")
                        continue
                    for each in results:
                        p_count[each['property']['value'].split("/")[-1]] += 1

                    for key, val in p_count.items():
                        if float(val) / len(unique_qnodes) >= search_threshold:
                            p_nodes_needed.append(key)
                    wikidata_search_result = {"p_nodes_needed": p_nodes_needed,
                                              "target_q_node_column_name": supplied_dataframe.columns[each_column]}
                    wikidata_results.append(DatamartSearchResult(search_result=wikidata_search_result,
                                                                 supplied_data=supplied_data,
                                                                 query_json=query,
                                                                 search_type="wikidata")
                                            )
            return wikidata_results

        except:
            print("Searching with wiki data failed")
            traceback.print_exc()
        finally:
            return wikidata_results


class DatamartMetadata:
    """
    This class represents the metadata associated with search results.
    """
    def __init__(self, title: str, description: str):
        self.title = title
        self.description = description

    def get_columns(self) -> typing.List[str]:
        """
        :return: a list of strings representing the names of columns in the dataset that can be downloaded from a
        search result.
        """
        pass

    def get_detailed_metadata(self) -> dict:
        """
        Datamart will use a program-wide metadata standard, currently computed by the Harvard team, using contributions
        from other performers. This method returns the standard metadata for a dataset.
        :return: a dict with the standard metadata.
        """
        pass

    def get_datamart_specific_metadata(self) -> dict:
        """
        A datamart implementation may compute additional metadata beyond the program-wide standard metadata.
        :return: a dict with the datamart-specific metadata.
        """
        pass


class DatamartSearchResult:
    """
    This class represents the search results of a datamart search.
    Different datamarts will provide different implementations of this class.

    Attributes
    ----------
    join_hints: typing.List[D3MAugmentSpec]
        Hints for joining supplied data with datamart data

    """
    def __init__(self, search_result, supplied_data, query_json, search_type):
        self._logger = logging.getLogger(__name__)
        self.search_result = search_result
        if "_score" in self.search_result:
            self.score = self.search_result["_score"]
        if "_source" in self.search_result:
            self.metadata = self.search_result["_source"]
        self.supplied_data = supplied_data

        if type(supplied_data) is d3m_Dataset:
            self.res_id, self.supplied_dataframe = d3m_utils.get_tabular_resource(dataset=supplied_data, resource_id=None)
            self.selector_base_type = "ds"
        elif type(supplied_data) is d3m_DataFrame:
            self.supplied_dataframe = supplied_data
            self.selector_base_type = "df"

        self.query_json = query_json
        self.search_type = search_type
        self.pairs = None
        self._res_id = None  # only used for input is Dataset
        self.join_pairs = None

    def display(self) -> pd.DataFrame:
        """
        function used to see what found inside this search result class in a human vision
        :return: a pandas DataFrame
        """
        if self.search_type == "wikidata":
            column_names = []
            for each in self.search_result["p_nodes_needed"]:
                each_name = self._get_node_name(each)
                column_names.append(each_name)
            column_names = ", ".join(column_names)
            required_variable = []
            required_variable.append(self.search_result["target_q_node_column_name"])
            result = pd.DataFrame({"title": "wikidata search result for " \
                                           + self.search_result["target_q_node_column_name"], \
                                   "columns": column_names, "join columns": required_variable}, index=[0])

        elif self.search_type == "general":
            title = self.search_result['_source']['title']
            column_names = []
            required_variable = []
            for each in self.query_json['required_variables']:
                required_variable.append(each['names'])

            for each in self.search_result['_source']['variables']:
                each_name = each['name']
                column_names.append(each_name)
            column_names = ", ".join(column_names)
            result = pd.DataFrame({"title": title, "columns": column_names, "join columns": required_variable}, index=[0])

        return result

    def download(self, supplied_data: typing.Union[d3m_Dataset, d3m_DataFrame], generate_metadata=True, return_format="ds") \
            -> typing.Union[d3m_Dataset, d3m_DataFrame]:
        """
        download the dataset or dataFrame (depending on the input type) and corresponding metadata information of
        search result everytime call download, the DataFrame will have the exact same columns in same order
        """
        if self.search_type=="general":
            return_df = self.download_general(supplied_data, generate_metadata, return_format)
        elif self.search_type=="wikidata":
            return_df = self.download_wikidata(supplied_data, generate_metadata, return_format)

        return return_df

    def download_general(self, supplied_data: typing.Union[d3m_Dataset, d3m_DataFrame]=None, generate_metadata=True,
                         return_format="ds", augment_resource_id = AUGMENT_RESOURCE_ID) -> typing.Union[d3m_Dataset, d3m_DataFrame]:
        """
        Specified download function for general datamart Datasets
        :param supplied_data: given supplied data
        :param generate_metadata: whether need to genreate the metadata or not
        :return: a dataset or a dataframe depending on the input
        """
        if type(supplied_data) is d3m_Dataset:
            self._res_id, self.supplied_dataframe = d3m_utils.get_tabular_resource(dataset=supplied_data, resource_id=None)
        else:
            self.supplied_dataframe = supplied_data

        if self.join_pairs is None:
            candidate_join_column_pairs = self.get_join_hints()
        else:
            candidate_join_column_pairs = self.join_pairs

        if len(candidate_join_column_pairs) > 1:
            print("[WARN]: multiple joining column pairs found")
        join_pairs_result = []
        candidate_join_column_scores = []

        # start finding pairs
        if supplied_data is None:
            supplied_data = self.supplied_dataframe
        left_df = copy.deepcopy(self.supplied_dataframe)
        right_metadata = self.search_result['_source']
        right_df = Utils.materialize(metadata=self.metadata)
        left_metadata = Utils.generate_metadata_from_dataframe(data=left_df, original_meta=None)

        # generate the pairs for each join_column_pairs
        for each_pair in candidate_join_column_pairs:
            left_columns = each_pair.left_columns
            right_columns = each_pair.right_columns
            try:
                # Only profile the joining columns, otherwise it will be too slow:
                left_metadata = Utils.calculate_dsbox_features(data=left_df, metadata=left_metadata,
                                                               selected_columns=set(left_columns))

                right_metadata = Utils.calculate_dsbox_features(data=right_df, metadata=right_metadata,
                                                                selected_columns=set(right_columns))
                # update with implicit_variable on the user supplied dataset
                if left_metadata.get('implicit_variables'):
                    Utils.append_columns_for_implicit_variables_and_add_meta(left_metadata, left_df)

                print(" - start getting pairs for", each_pair.to_str_format())

                result, self.pairs = RLTKJoiner.find_pair(left_df=left_df, right_df=right_df,
                                                          left_columns=[left_columns], right_columns=[right_columns],
                                                          left_metadata=left_metadata, right_metadata=right_metadata)

                join_pairs_result.append(result)
                # TODO: figure out some way to compute the joining quality
                candidate_join_column_scores.append(100)
            except:
                print("failed when getting pairs for", each_pair)
                traceback.print_exc()

        # choose the best joining results
        all_results = []
        for i in range(len(join_pairs_result)):
            each_result = (candidate_join_column_pairs[i], candidate_join_column_scores[i], join_pairs_result[i])
            all_results.append(each_result)

        all_results.sort(key=lambda x: x[1], reverse=True)
        if len(all_results) == 0:
            raise ValueError("[ERROR] Failed to get pairs!")

        if return_format == "ds":
            return_df = d3m_DataFrame(all_results[0][2], generate_metadata=False)
            resources = {augment_resource_id: return_df}
            return_result = d3m_Dataset(resources=resources, generate_metadata=False)
            if generate_metadata:
                metadata_shape_part_dict = self._generate_metadata_shape_part(value=return_result, selector=())
                for each_selector, each_metadata in metadata_shape_part_dict.items():
                    return_result.metadata = return_result.metadata.update(selector=each_selector, metadata=each_metadata)
                return_result.metadata = self._generate_metadata_column_part_for_general(return_result.metadata, return_format, augment_resource_id)

        elif return_format == "df":
            return_result = d3m_DataFrame(all_results[0][2], generate_metadata=False)
            if generate_metadata:
                metadata_shape_part_dict = self._generate_metadata_shape_part(value=return_result, selector=())
                for each_selector, each_metadata in metadata_shape_part_dict.items():
                    return_result.metadata = return_result.metadata.update(selector=each_selector, metadata=each_metadata)
                return_result.metadata = self._generate_metadata_column_part_for_general(return_result.metadata, return_format, augment_resource_id=None)

        else:
            raise ValueError("Invalid return format was given")

        return return_result

    def _generate_metadata_shape_part(self, value, selector) -> dict:
        """
        recursively generate all metadata for shape part, return a dict
        :param value:
        :param selector:
        :return:
        """
        generated_metadata: dict = {}
        generated_metadata['schema'] = CONTAINER_SCHEMA_VERSION
        if isinstance(value, d3m_Dataset):  # type: ignore
            generated_metadata['dimension'] = {
                'name': 'resources',
                'semantic_types': ['https://metadata.datadrivendiscovery.org/types/DatasetResource'],
                'length': len(value),
            }

            metadata_dict = collections.OrderedDict([(selector, generated_metadata)])

            for k, v in value.items():
                metadata_dict.update(self._generate_metadata_shape_part(v, selector + (k,)))

            # It is unlikely that metadata is equal across dataset resources, so we do not try to compact metadata here.

            return metadata_dict

        if isinstance(value, d3m_DataFrame):  # type: ignore
            generated_metadata['semantic_types'] = ['https://metadata.datadrivendiscovery.org/types/Table']

            generated_metadata['dimension'] = {
                'name': 'rows',
                'semantic_types': ['https://metadata.datadrivendiscovery.org/types/TabularRow'],
                'length': value.shape[0],
            }

            metadata_dict = collections.OrderedDict([(selector, generated_metadata)])

            # Reusing the variable for next dimension.
            generated_metadata = {
                'dimension': {
                    'name': 'columns',
                    'semantic_types': ['https://metadata.datadrivendiscovery.org/types/TabularColumn'],
                    'length': value.shape[1],
                },
            }

            selector_all_rows = selector + (ALL_ELEMENTS,)
            metadata_dict[selector_all_rows] = generated_metadata
            return metadata_dict

    def _generate_metadata_column_part_for_general(self, metadata_return, return_format, augment_resource_id) -> DataMetadata:
        """
        Inner function used to generate metadata for general search
        """
        # part for adding the whole dataset/ dataframe's metadata


        # part for adding each column's metadata
        for i, each_metadata in enumerate(self.metadata['variables']):
            if return_format == "ds":
                metadata_selector = (augment_resource_id, ALL_ELEMENTS, i)
            elif return_format == "df":
                metadata_selector = (ALL_ELEMENTS, i)
            structural_type = each_metadata["description"].split("dtype: ")[-1]
            if "int" in structural_type:
                structural_type = int
            elif "float" in structural_type:
                structural_type = float
            else:
                structural_type = str
            metadata_each_column = {"name": each_metadata["name"], "structural_type": structural_type,
                                    'semantic_types': ('https://metadata.datadrivendiscovery.org/types/Attribute',)}
            metadata_return = metadata_return.update(metadata=metadata_each_column, selector=metadata_selector)

        if return_format == "ds":
            metadata_selector = (augment_resource_id, ALL_ELEMENTS, i + 1)
        elif return_format == "df":
            metadata_selector = (ALL_ELEMENTS, i + 1)
        metadata_joining_pairs = {"name": "joining_pairs", "structural_type": typing.List[int],
                                  'semantic_types': ("http://schema.org/Integer",)}
        metadata_return = metadata_return.update(metadata=metadata_joining_pairs, selector=metadata_selector)

        return metadata_return

    def download_wikidata(self, supplied_data: typing.Union[d3m_Dataset, d3m_DataFrame], generate_metadata=True, return_format="ds",augment_resource_id=AUGMENT_RESOURCE_ID) -> typing.Union[d3m_Dataset, d3m_DataFrame]:
        """
        :param supplied_data: input DataFrame
        :param generate_metadata: control whether to automatically generate metadata of the return DataFrame or not
        :return: return_df: the materialized wikidata d3m_DataFrame,
                            with corresponding pairing information to original_data at last column
        """
        # prepare the query
        p_nodes_needed = self.search_result["p_nodes_needed"]
        target_q_node_column_name = self.search_result["target_q_node_column_name"]
        if type(supplied_data) is d3m_DataFrame:
            self.supplied_dataframe = copy.deepcopy(supplied_data)
        elif type(supplied_data) is d3m_Dataset:
            self._res_id, supplied_dataframe = d3m_utils.get_tabular_resource(dataset=supplied_data,
                                                                                  resource_id=None)
            self.supplied_dataframe = copy.deepcopy(supplied_dataframe)

        q_node_column_number = self.supplied_dataframe.columns.tolist().index(target_q_node_column_name)
        q_nodes_list = set(self.supplied_dataframe.iloc[:, q_node_column_number].tolist())
        q_nodes_query = ""
        p_nodes_query_part = ""
        p_nodes_optional_part = ""
        special_request_part = ""
        # q_nodes_list = q_nodes_list[:30]
        for each in q_nodes_list:
            if each != "N/A":
                q_nodes_query += "(wd:" + each + ") \n"
        for each in p_nodes_needed:
            if each not in P_NODE_IGNORE_LIST:
                p_nodes_query_part += " ?" + each
                p_nodes_optional_part += "  OPTIONAL { ?q wdt:" + each + " ?" + each + "}\n"
            if each in SPECIAL_REQUEST_FOR_P_NODE:
                special_request_part += SPECIAL_REQUEST_FOR_P_NODE[each] + "\n"

        sparql_query = "SELECT DISTINCT ?q " + p_nodes_query_part + \
                       "WHERE \n{\n  VALUES (?q) { \n " + q_nodes_query + "}\n" + \
                       p_nodes_optional_part + special_request_part + "}\n"
        import pdb
        pdb.set_trace()
        return_df = d3m_DataFrame()
        try:
            sparql = SPARQLWrapper(WIKIDATA_QUERY_SERVER)
            sparql.setQuery(sparql_query)
            sparql.setReturnFormat(JSON)
            sparql.setMethod(POST)

            sparql.setRequestMethod(URLENCODED)
            results = sparql.query().convert()
        except:
            print("Getting query of wiki data failed!")
            return return_df

        semantic_types_dict = {
            "q_node": ("http://schema.org/Text", 'https://metadata.datadrivendiscovery.org/types/PrimaryKey')}

        for result in results["results"]["bindings"]:
            each_result = {}
            q_node_name = result.pop("q")["value"].split("/")[-1]
            each_result["q_node"] = q_node_name
            for p_name, p_val in result.items():
                each_result[p_name] = p_val["value"]
                # only do this part if generate_metadata is required
                if p_name not in semantic_types_dict:
                    if "datatype" in p_val.keys():
                        semantic_types_dict[p_name] = (
                            self._get_semantic_type(p_val["datatype"]),
                            'https://metadata.datadrivendiscovery.org/types/Attribute')
                    else:
                        semantic_types_dict[p_name] = (
                            "http://schema.org/Text", 'https://metadata.datadrivendiscovery.org/types/Attribute')

            return_df = return_df.append(each_result, ignore_index=True)

        p_name_dict = {"q_node": "q_node"}
        for each in return_df.columns.tolist():
            if each.lower().startswith("p") or each.lower().startswith("c"):
                p_name_dict[each] = self._get_node_name(each)

        # use rltk joiner to find the joining pairs
        joiner = RLTKJoiner_new()
        joiner.set_join_target_column_names((self.supplied_dataframe.columns[q_node_column_number], "q_node"))
        result, self.pairs = joiner.find_pair(left_df=self.supplied_dataframe, right_df=return_df)

        # if this condition is true, it means "id" column was added which should not be here
        if return_df.shape[1] == len(p_name_dict) + 2 and "id" in return_df.columns:
            return_df = return_df.drop(columns=["id"])

        metadata_new = DataMetadata()
        self.metadata = {}

        # add remained attributes metadata
        for each_column in range(0, return_df.shape[1] - 1):
            current_column_name = p_name_dict[return_df.columns[each_column]]
            metadata_selector = (ALL_ELEMENTS, each_column)
            # here we do not modify the original data, we just add an extra "expected_semantic_types" to metadata
            metadata_each_column = {"name": current_column_name, "structural_type": str,
                                    'semantic_types': semantic_types_dict[return_df.columns[each_column]]}
            self.metadata[current_column_name] = metadata_each_column
            if generate_metadata:
                metadata_new = metadata_new.update(metadata=metadata_each_column, selector=metadata_selector)

        # special for joining_pairs column
        metadata_selector = (ALL_ELEMENTS, return_df.shape[1])
        metadata_joining_pairs = {"name": "joining_pairs", "structural_type": typing.List[int],
                                  'semantic_types': ("http://schema.org/Integer",)}
        if generate_metadata:
            metadata_new = metadata_new.update(metadata=metadata_joining_pairs, selector=metadata_selector)

        # start adding shape metadata for dataset
        if return_format == "ds":
            return_df = d3m_DataFrame(return_df, generate_metadata=False)
            return_df = return_df.rename(columns=p_name_dict)
            resources = {augment_resource_id: return_df}
            return_result = d3m_Dataset(resources=resources, generate_metadata=False)
            if generate_metadata:
                return_result.metadata = metadata_new
                metadata_shape_part_dict = self._generate_metadata_shape_part(value=return_result, selector=())
                for each_selector, each_metadata in metadata_shape_part_dict.items():
                    return_result.metadata = return_result.metadata.update(selector=each_selector,
                                                                           metadata=each_metadata)
            # update column names to be property names instead of number

        elif return_format == "df":
            return_result = d3m_DataFrame(return_df, generate_metadata=False)
            return_result = return_result.rename(columns=p_name_dict)
            if generate_metadata:
                return_result.metadata = metadata_new
                metadata_shape_part_dict = self._generate_metadata_shape_part(value=return_result, selector=())
                for each_selector, each_metadata in metadata_shape_part_dict.items():
                    return_result.metadata = return_result.metadata.update(selector=each_selector,
                                                                           metadata=each_metadata)
        return return_result

    def _get_node_name(self, node_code):
        """
        Inner function used to get the properties(P nodes) names with given P node
        :param node_code: a str indicate the P node (e.g. "P123")
        :return: a str indicate the P node label (e.g. "inception")
        """
        sparql_query = "SELECT DISTINCT ?x WHERE \n { \n" + \
                       "wd:" + node_code + " rdfs:label ?x .\n FILTER(LANG(?x) = 'en') \n} "
        try:
            sparql = SPARQLWrapper(WIKIDATA_QUERY_SERVER)
            sparql.setQuery(sparql_query)
            sparql.setReturnFormat(JSON)
            sparql.setMethod(POST)
            sparql.setRequestMethod(URLENCODED)
            results = sparql.query().convert()
            return results['results']['bindings'][0]['x']['value']
        except:
            print("Getting name of node " + node_code + " failed!")
            return node_code

    def _get_semantic_type(self, datatype: str):
        """
        Inner function used to transfer the wikidata semantic type to D3M semantic type
        :param datatype: a str indicate the semantic type adapted from wikidata
        :return: a str indicate the semantic type for D3M
        """
        special_type_dict = {"http://www.w3.org/2001/XMLSchema#dateTime": "http://schema.org/DateTime",
                             "http://www.w3.org/2001/XMLSchema#decimal": "http://schema.org/Float",
                             "http://www.opengis.net/ont/geosparql#wktLiteral": "https://metadata.datadrivendiscovery.org/types/Location"}
        default_type = "http://schema.org/Text"
        if datatype in special_type_dict:
            return special_type_dict[datatype]
        else:
            print("not seen type : ", datatype)
            return default_type

    def augment(self, supplied_data, generate_metadata=True, augment_resource_id=AUGMENT_RESOURCE_ID):
        """
        download and join using the D3mJoinSpec from get_join_hints()
        """
        if type(supplied_data) is d3m_DataFrame:
            return self._augment(supplied_data=supplied_data, generate_metadata=generate_metadata, return_format="df", augment_resource_id=augment_resource_id)
        elif type(supplied_data) is d3m_Dataset:
            self._res_id, self.supplied_data = d3m_utils.get_tabular_resource(dataset=supplied_data, resource_id=None, has_hyperparameter=False)
            res = self._augment(supplied_data=supplied_data, generate_metadata=generate_metadata, return_format="ds", augment_resource_id=augment_resource_id)
            return res

    def _augment(self, supplied_data, generate_metadata=True, return_format="ds", augment_resource_id=AUGMENT_RESOURCE_ID):
        """
        download and join using the D3mJoinSpec from get_join_hints()
        """
        if type(supplied_data) is d3m_Dataset:
            supplied_data_df = supplied_data[self._res_id]
        else:
            supplied_data_df = supplied_data

        download_result = self.download(supplied_data=supplied_data_df, generate_metadata=False, return_format="df")
        download_result = download_result.drop(columns=['joining_pairs'])
        df_joined = pd.DataFrame()

        column_names_to_join = None
        for r1, r2 in self.pairs:
            left_res = supplied_data_df.loc[int(r1)]
            right_res = download_result.loc[int(r2)]
            if column_names_to_join is None:
                column_names_to_join = right_res.index.difference(left_res.index)
                matched_rows = right_res.index.intersection(left_res.index)
                columns_new = left_res.index.tolist()
                columns_new.extend(column_names_to_join.tolist())
            new = pd.concat([left_res, right_res[column_names_to_join]])
            df_joined = df_joined.append(new, ignore_index=True)
        # ensure that the original dataframe columns are at the first left part
        df_joined = df_joined[columns_new]
        # if search with wikidata, we can remove duplicate Q node column

        if self.search_type == "wikidata":
            df_joined = df_joined.drop(columns=['q_node'])

        if 'id' in df_joined.columns:
            df_joined = df_joined.drop(columns=['id'])

        if generate_metadata:
            columns_all = list(df_joined.columns)
            if 'd3mIndex' in df_joined.columns:
                oldindex = columns_all.index('d3mIndex')
                columns_all.insert(0, columns_all.pop(oldindex))
            else:
                self._logger.warning("No d3mIndex column found after data-mart augment!!!")
            df_joined = df_joined[columns_all]

        # start adding column metadata for dataset
        if generate_metadata:
            metadata_dict_left = {}
            metadata_dict_right = {}
            if self.search_type == "general":
                for each in self.metadata['variables']:
                    description = each['description']
                    dtype = description.split("dtype: ")[-1]
                    if "float" in dtype:
                        semantic_types = (
                            "http://schema.org/Float",
                            "https://metadata.datadrivendiscovery.org/types/Attribute"
                        )
                    elif "int" in dtype:
                        semantic_types = (
                            "http://schema.org/Integer",
                            "https://metadata.datadrivendiscovery.org/types/Attribute"
                        )
                    else:
                        semantic_types = (
                            "https://metadata.datadrivendiscovery.org/types/CategoricalData",
                            "https://metadata.datadrivendiscovery.org/types/Attribute"
                        )
                    each_meta = {
                        "name": each['name'],
                        "structural_type": str,
                        "semantic_types": semantic_types,
                        "description": description
                    }
                    metadata_dict_right[each['name']] = frozendict.FrozenOrderedDict(each_meta)
            else:
                metadata_dict_right = self.metadata

            if return_format == "df":
                left_df_column_legth = supplied_data.metadata.query((metadata_base.ALL_ELEMENTS,))['dimension'][
                    'length']
            elif return_format == "ds":
                left_df_column_legth = supplied_data.metadata.query((self._res_id, metadata_base.ALL_ELEMENTS,))['dimension']['length']

            # add the original metadata
            for i in range(left_df_column_legth):
                if return_format == "df":
                    each_selector = (ALL_ELEMENTS, i)
                elif return_format == "ds":
                    each_selector = (self._res_id, ALL_ELEMENTS, i)
                each_column_meta = supplied_data.metadata.query(each_selector)
                metadata_dict_left[each_column_meta['name']] = each_column_meta

            metadata_new = metadata_base.DataMetadata()
            metadata_old = copy.copy(supplied_data.metadata)

            new_column_names_list = list(df_joined.columns)
            # update each column's metadata
            for i, current_column_name in enumerate(new_column_names_list):
                if return_format == "df":
                    each_selector = (metadata_base.ALL_ELEMENTS, i)
                elif return_format == "ds":
                    each_selector = (augment_resource_id, ALL_ELEMENTS, i)
                if current_column_name in metadata_dict_left:
                    new_metadata_i = metadata_dict_left[current_column_name]
                else:
                    new_metadata_i = metadata_dict_right[current_column_name]
                metadata_new = metadata_new.update(each_selector, new_metadata_i)

            # start adding shape metadata for dataset
            if return_format == "ds":
                return_df = d3m_DataFrame(df_joined, generate_metadata=False)
                resources = {augment_resource_id: return_df}
                return_result = d3m_Dataset(resources=resources, generate_metadata=False)
                if generate_metadata:
                    return_result.metadata = metadata_new
                    metadata_shape_part_dict = self._generate_metadata_shape_part(value=return_result, selector=())
                    for each_selector, each_metadata in metadata_shape_part_dict.items():
                        return_result.metadata = return_result.metadata.update(selector=each_selector,
                                                                               metadata=each_metadata)
            elif return_format == "df":
                return_result = d3m_DataFrame(df_joined, generate_metadata=False)
                if generate_metadata:
                    return_result.metadata = metadata_new
                    metadata_shape_part_dict = self._generate_metadata_shape_part(value=return_result, selector=())
                    for each_selector, each_metadata in metadata_shape_part_dict.items():
                        return_result.metadata = return_result.metadata.update(selector=each_selector,
                                                                               metadata=each_metadata)

            return return_result

    def get_score(self) -> float:
        return self.score

    def get_metadata(self) -> dict:
        return self.metadata

    def set_join_pairs(self, join_pairs: typing.List[D3MJoinSpec]) -> None:
        """
        manually set up the join pairs
        :param join_pairs: user specified D3MJoinSpec
        :return:
        """
        self.join_pairs = join_pairs

    def get_join_hints(self, supplied_data=None) -> typing.List[D3MJoinSpec]:
        """
        Returns hints for joining supplied data with the data that can be downloaded using this search result.
        In the typical scenario, the hints are based on supplied data that was provided when search was called.

        The optional supplied_data argument enables the caller to request recomputation of join hints for specific data.

        :return: a list of join hints. Note that datamart is encouraged to return join hints but not required to do so.
        """
        if not supplied_data:
            supplied_data = self.supplied_dataframe
        if self.search_type == "general":
            inner_hits = self.search_result.get('inner_hits', {})
            results = []
            used = set()
            for key_path, outer_hits in inner_hits.items():
                vars_type, index, ent_type = key_path.split('.')
                if vars_type != 'required_variables':
                    continue
                left_index = []
                right_index = []
                index = int(index)
                if ent_type == JSONQueryManager.DATAFRAME_COLUMNS:
                    if self.query_json[vars_type][index].get('index'):
                        left_index = self.query_json[vars_type][index].get('index')
                    elif self.query_json[vars_type][index].get('names'):
                        left_index = [supplied_data.columns.tolist().index(idx)
                                      for idx in self.query_json[vars_type][index].get('names')]

                    inner_hits = outer_hits.get('hits', {})
                    hits_list = inner_hits.get('hits')
                    if hits_list:
                        for hit in hits_list:
                            offset = hit['_nested']['offset']
                            if offset not in used:
                                right_index.append(offset)
                                used.add(offset)
                                break

                if left_index and right_index:
                    each_result = D3MJoinSpec(left_columns=left_index, right_columns=right_index)
                    results.append(each_result)

            return results

    @classmethod
    def construct(cls, serialization: dict) -> DatamartSearchResult:
        """
        Take into the serilized input and reconsctruct a "DatamartSearchResult"
        """
        load_result = DatamartSearchResult(search_result=serialization["search_result"],
                                           supplied_data=None,
                                           query_json=serialization["query_json"],
                                           search_type=serialization["search_type"])
        return load_result

    def serialize(self) -> dict:
        output = dict()
        output["search_result"] = self.search_result
        output["query_json"] = self.query_json
        output["search_type"] = self.search_type
        return output


class D3MJoinSpec:
    """
    A join spec specifies a possible way to join a left datasets with a right dataset. The spec assumes that it may
    be necessary to use several columns in each datasets to produce a key or fingerprint that is useful for joining
    datasets. The spec consists of two lists of column identifiers or names (left_columns, left_column_names and
    right_columns, right_column_names).

    In the simplest case, both left and right are singleton lists, and the expectation is that an appropriate
    matching function exists to adequately join the datasets. In some cases equality may be an appropriate matching
    function, and in some cases fuzz matching is required. The spec join spec does not specify the matching function.

    In more complex cases, one or both left and right lists contain several elements. For example, the left list
    may contain columns for "city", "state" and "country" and the right dataset contains an "address" column. The join
    spec pairs up ["city", "state", "country"] with ["address"], but does not specify how the matching should be done
    e.g., combine the city/state/country columns into a single column, or split the address into several columns.
    """
    def __init__(self, left_columns: typing.List[typing.List[int]], right_columns: typing.List[typing.List[int]],
                 left_resource_id: str=None, right_resource_id: str=None):
        self.left_resource_id = left_resource_id
        self.right_resource_id = right_resource_id
        self.left_columns = left_columns
        self.right_columns = right_columns
        # we can have list of the joining column pairs
        # each list inside left_columns/right_columns is a candidate joining column for that dataFrame
        # each candidate joining column can also have multiple columns

    def to_str_format(self):
        return "[ (" + (self.left_resource_id or "") + ", " + str(self.left_columns) + ") , (" + \
               (self.right_resource_id or "") + ", " + str(self.right_columns) + ") ]"


class TemporalGranularity(enum.Enum):
    YEAR = 1
    MONTH = 2
    DAY = 3
    HOUR = 4
    SECOND = 5


class GeospatialGranularity(enum.Enum):
    COUNTRY = 1
    STATE = 2
    COUNTY = 3
    CITY = 4
    POSTALCODE = 5


class DatamartVariable:
    pass


class DatamartQuery:
    def __init__(self, about: str=None, required_variables: typing.List[DatamartVariable]=None,
                 desired_variables: typing.List[DatamartVariable]=None):
        self.about = about
        self.required_variables = required_variables
        self.desired_variables = desired_variables

    def to_json(self):
        """
        function used to transform the Query to json format that can used on elastic search
        :return:
        """
        search_query = dict()
        if self.about is not None:
            search_query["dataset"] = {
                    "about": self.about
                }
        if self.required_variables is not None:
            search_query["required_variables"] = self.required_variables

        return search_query

class NamedEntityVariable(DatamartVariable):
    """
    Describes columns containing named enitities.

    Parameters
    ----------
    entities: List[str]
        List of strings that should be contained in the matched dataset column.
    """
    def __init__(self, entities: typing.List[str]):
        self.entities = entities


class TemporalVariable(DatamartVariable):
    """
    Describes columns containing temporal information.

    Parameters
    ----------
    start: datetime
        Requested datetime should be equal or older than this datetime
    end: datetime
        Requested datatime should be equal or less than this datetime
    granularity: TemporalGranularity
        Requested datetimes are well matched with the requested granularity
    """
    def __init__(self, start: datetime.datetime, end: datetime.datetime, granularity: TemporalGranularity):
        pass


class GeospatialVariable(DatamartVariable):
    """
    Describes columns containing geospatial information.

    Parameters
    ----------
    lat1: float
        The latitude of the first point
    long1: float
        The longitude of the first point
    lat2: float
        The latitude of the second point
    long2: float
        The longitude of the second point
    granularity: GeospatialGranularity
        Requested geosptial values are well matched with the requested granularity
    """
    def __init__(self, lat1: float, long1: float, lat2: float, long2: float, granularity: GeospatialGranularity):
        pass



from setuptools import setup, find_packages

with open('requirements.txt', 'r') as f:
    install_requires = list()
    dependency_links = list()
    for line in f:
        re = line.strip()
        if re:
            install_requires.append(re)

setup(name='datamart_isi',
      version='2.1',
      description='USC ISI implementation of D3M Datamart API',
      long_description='USC ISI implementation of D3M Datamart API',
      author='USC ISI',
      url='https://github.com/usc-isi-i2/datamart-userend/tree/d3m',
      packages=find_packages(),

      install_requires=install_requires,
      classifiers=[
          "Programming Language :: Python :: 3",
          "License :: OSI Approved :: MIT License",
          "Operating System :: OS Independent",
      ])



from datamart_isi.materializers.parsers.csv_parser import CSVParser
from datamart_isi.materializers.parsers.excel_parser import ExcelParser
from datamart_isi.materializers.parsers.json_parser import JSONParser
from datamart_isi.materializers.parsers.html_parser import HTMLParser
from pkgutil import extend_path
__path__ = extend_path(__path__, __name__)


from numpy import nan

from datamart_isi.materializers.parsers.parser_base import *


class ExcelParser(ParserBase):

    @staticmethod
    def encode_url(url):
        return url.replace(" ", "%20")

    def get_all(self, url: str) -> typing.List[pd.DataFrame]:
        return [_.dataframe for _ in self.parse(url)]

    def parse(self, url: str) -> typing.List[ParseResult]:
        """
        Retrive excel file from url and return Dataframe and metadata for every sheet

        Params:
            - url: (str)
        
        Returns:
            - dfs: (ParseResult) ParseResult obj containing dataframes and metadata dicts
        """
        url = self.encode_url(url)
        xl = pd.ExcelFile(url)
        dfs = []
        for idx, sheet in enumerate(xl.sheet_names):
            meta, start_row, skipfooter = self._parse_metadata(xl, sheet)
            df = xl.parse(header=start_row, skipfooter=skipfooter, sheet_name=sheet)
            df = df.dropna(how="all",axis=0)
            df = df.dropna(how="all",axis=1)
            dfs.append(ParseResult(
                df=df,
                index=idx,
                name=sheet,
                metadata=meta
            ))
        return dfs
    
    def _parse_metadata(self, xl, sheetname):
        """
        Read excel sheet and parse metadata from it

        Params:
            - xl: (ExcelFile) pandas ExcelFile objects
            - sheetname: (str) name of the sheet to parse

        Returns:
            - metadata: (dict) metadata extracted from the file
            - start_row: (int) start index of data
            - skipfooter: (int) index of data starting from the end of table

        """
        df = xl.parse(sheet_name=sheetname, index_col=None, header=None)
        df = df.replace(' ', nan)
        df = df.dropna(how='all', axis=1)
        rows, cols = df.shape
        metadata = {}
        start_row = 0
        # start from beginning
        # find starting metadata
        for row in range(rows):
            if df.iloc[row].isnull().sum() > cols / 2:
                # is metadata
                item = df.iloc[row].dropna().tolist()
                if len(item) > 0:
                    metadata[row] = item
            else:
                start_row = row
                break
        # start from end
        # find footer metadata
        end_row = rows
        for row in range(rows - 1, -1, -1):
            if df.iloc[row].isnull().sum() > cols / 2:
                # is metadata
                item = df.iloc[row].dropna().tolist()
                if len(item) > 0:
                    metadata[row] = item
            else:

                end_row = row + 1
                break
        skipfooter = rows - end_row
        return metadata, start_row, skipfooter


import pandas as pd  # type: ignore
import numpy as np  # type: ignore
from collections import Counter
from builtins import filter


def ordered_dict2(column, k):
    unique, counts = np.unique(column, return_counts=True)
    d = dict(zip(unique, counts))
    dlist = []
    for k, v in Counter(d).most_common(k):
        e = {'name': k, 'count': v}
        dlist.append(e)
    return dlist


def ordered_dict(column, k):
    # d = column.value_counts()[:k].to_dict()
    d = column.value_counts().head(k).to_dict()
    dlist = []
    for k, v in sorted(d.items(), key=lambda x: x[1], reverse=True):
        e = {'name': k, 'count': v}
        dlist.append(e)
    return dlist


def tryConvert(cell):
    """
    convert a cell, if possible, to its supposed type(int, float, string)
    note: type of NaN cell is float
    """
    try:
        return int(cell)
    except ValueError as TypeError:
        try:
            return float(cell)
        except ValueError as TypeError:
            return cell


def is_Decimal_Number(s):
    try:
        float(s)
        return True
    except:
        return False


def numerical_stats(feature, column, num_nonblank, feature_list):
    """
    calculates numerical statistics
    """
    ## hyper-parameter ##
    sigma = 3
    ## =============== ##
    stats = column.describe()
    if ("number_of_numeric_values" in feature_list):
        feature["number_of_numeric_values"] = int(stats["count"])
    if ("ratio_of_numeric_values" in feature_list):
        feature["ratio_of_numeric_values"] = stats["count"] / num_nonblank
    # feature["number_mean"] = stats["mean"]
    # feature["number_std"] = stats["std"]
    if stats["count"] == 1: feature["number_std"] = 0
    # feature["number_quartile_1"] = stats["25%"]
    # feature["number_quartile_2"] = stats["50%"]
    # feature["number_quartile_3"] = stats["75%"]
    if ("number_of_outlier_numeric_values" in feature_list):
        outlier = column[(np.abs(column - stats["mean"]) > (sigma * stats["std"]))]
        feature["number_of_outlier_numeric_values"] = outlier.count()
    if ("number_of_positive_numeric_values" in feature_list):
        feature["number_of_positive_numeric_values"] = column[column > 0].count()
    if ("number_of_negative_numeric_values" in feature_list):
        feature["number_of_negative_numeric_values"] = column[column < 0].count()
    if ("number_of_numeric_values_equal_0" in feature_list):
        feature["number_of_numeric_values_equal_0"] = column[column == 0].count()
    if ("number_of_numeric_values_equal_1" in feature_list):
        feature["number_of_numeric_values_equal_1"] = column[column == 1].count()
    if ("number_of_numeric_values_equal_-1" in feature_list):
        feature["number_of_numeric_values_equal_-1"] = column[column == -1].count()

    if ("target_values" in feature_list):
        feature["target_values"] = {'mean': stats['mean'],
                                    'std': stats['std'],
                                    'median': stats['50%'],
                                    'quartile_1': stats['25%'],
                                    'quartile_3': stats['75%']}


def compute_numerics(column, feature, feature_list):
    """
    computes numerical features of the column:
    # of integers/ decimal(float only)/ nonblank values in the column
    statistics of int/decimal/numerics
    """

    # if one of them is specified, then go next
    if (("number_of_numeric_values" in feature_list) or
            ("ratio_of_numeric_values" in feature_list) or
            ("number_std" in feature_list) or
            ("number_of_outlier_numeric_values" in feature_list) or
            ("number_of_negative_numeric_values" in feature_list) or
            ("number_of_positive_numeric_values" in feature_list) or
            ("number_of_numeric_values_equal_0" in feature_list) or
            ("number_of_numeric_values_equal_1" in feature_list) or
            ("number_of_numeric_values_equal_-1" in feature_list) or
            ("target_values" in feature_list)):

        cnt = column.count()
        if column.dtype.kind in np.typecodes['AllInteger'] + 'uf' and cnt > 0:
            numerical_stats(feature, column, cnt, feature_list)
        else:
            convert = lambda v: tryConvert(v)
            col = column.apply(convert, convert_dtype=False)

            col_nonblank = col.dropna()
            col_num = pd.Series([e for e in col_nonblank if
                                 type(e) == int or type(e) == np.int64 or type(e) == float or type(e) == np.float64])

            if col_num.count() > 0:
                numerical_stats(feature, col_num, cnt, feature_list)


def compute_common_numeric_tokens(column, feature, k):
    """
    compute top k frequent numerical tokens and their counts.
    tokens are integer or floats
    e.g. "123", "12.3"
    """
    col = np.asarray([token for lst in column.str.split().dropna() for token in lst])
    token = np.array(list(filter(lambda x: is_Decimal_Number(x), col)))
    if token.size:
        feature["most_common_numeric_tokens"] = ordered_dict2(token, k)


def compute_common_alphanumeric_tokens(column, feature, k):
    """
    compute top k frequent alphanumerical tokens and their counts.
    tokens only contain alphabets and/or numbers, decimals with points not included
    """
    col = np.asarray([token for lst in column.str.split().dropna() for token in lst])
    token = np.array(list(filter(lambda x: x.isalnum(), col)))
    if token.size:
        feature["most_common_alphanumeric_tokens"] = ordered_dict2(token, k)


def compute_common_values(column, feature, k):
    """
    compute top k frequent cell values and their counts.
    """
    if column.count() > 0:
        feature["most_common_raw_values"] = ordered_dict(column, k)


def compute_common_tokens(column, feature, k, feature_list):
    """
    compute top k frequent tokens and their counts.
    currently: tokens separated by white space
    at the same time, count on tokens which contain number(s)
    e.g. "$100", "60F", "123-456-7890"
    note: delimiter = " "
    """

    # if one of them specified, just compute all
    if (("most_common_tokens" in feature_list) or
            ("number_of_tokens_containing_numeric_char" in feature_list) or
            ("ratio_of_tokens_containing_numeric_char" in feature_list)):

        token = np.asarray([token for lst in column.str.split().dropna() for token in lst])
        if token.size:
            feature["most_common_tokens"] = ordered_dict2(token, k)
            cnt = sum([any(char.isdigit() for char in c) for c in token])
            if cnt > 0:
                feature["number_of_tokens_containing_numeric_char"] = cnt
                feature["ratio_of_tokens_containing_numeric_char"] = float(cnt) / token.size


def compute_common_tokens_by_puncs(column, feature, k, feature_list):
    """
    tokens seperated by all string.punctuation characters:
    '!"#$%&\'()*+,-./:;<=>?@[\\]^_`{|}~'
    """
    # if one of them specified, just compute all
    if (("most_common_tokens_split_by_punctuation" in feature_list) or
            ("number_of_distinct_tokens_split_by_punctuation" in feature_list) or
            ("ratio_of_distinct_tokens_split_by_punctuation" in feature_list) or
            ("number_of_tokens_split_by_punctuation_containing_numeric_char" in feature_list) or
            ("ratio_of_tokens_split_by_punctuation_containing_numeric_char" in feature_list)):

        col = column.dropna().values
        token_nested = [("".join((word if word.isalnum() else " ") for word in char).split()) for char in col]
        token = np.array([item for sublist in token_nested for item in sublist])
        if token.size:
            feature["most_common_tokens_split_by_punctuation"] = ordered_dict2(token, k)
            dist_cnt = np.unique(token).size
            feature["number_of_distinct_tokens_split_by_punctuation"] = dist_cnt
            feature["ratio_of_distinct_tokens_split_by_punctuation"] = float(dist_cnt) / token.size
            cnt = sum([any(char.isdigit() for char in c) for c in token])
            if cnt > 0:
                feature["number_of_tokens_split_by_punctuation_containing_numeric_char"] = cnt
                feature["ratio_of_tokens_split_by_punctuation_containing_numeric_char"] = float(cnt) / token.size


def compute_numeric_density(column, feature):
    """
    compute overall density of numeric characters in the column.
    """
    col = column.dropna().values
    if col.size:
        density = np.array([(sum(char.isdigit() for char in c), len(c)) for c in col])
        digit_total = density.sum(axis=0)
        feature["numeric_char_density"] = {'mean': float(digit_total[0]) / digit_total[1]}


def compute_contain_numeric_values(column, feature, feature_list):
    """
    caculate # and ratio of cells in the column which contains numbers.
    """
    # if one of them specified, just compute all
    if (("number_of_values_containing_numeric_char" in feature_list) or
            ("ratio_of_values_containing_numeric_char" in feature_list)):

        contain_digits = lambda x: any(char.isdigit() for char in x)
        cnt = column.dropna().apply(contain_digits).sum()
        if cnt > 0:
            feature["number_of_values_containing_numeric_char"] = cnt
            feature["ratio_of_values_containing_numeric_char"] = float(cnt) / column.count()



import pandas as pd


class ParseResult(object):
    def __init__(self, df: pd.DataFrame, index: int=None, name: str=None, metadata: dict=None):
        self._dataframe = df
        self._index = index
        self._name = name
        self._metadata = metadata

    @property
    def dataframe(self):
        return self._dataframe

    @property
    def index(self):
        return self._index

    @property
    def name(self):
        return self._name

    @property
    def metadata(self):
        return self._metadata



import numpy as np
import pandas as pd

from datamart_isi.profilers.helpers import feature_compute_hih as fc_hih, feature_compute_lfh as fc_lfh

computable_metafeatures = [
    'ratio_of_values_containing_numeric_char', 'ratio_of_numeric_values',
    'number_of_outlier_numeric_values', 'num_filename', 'number_of_tokens_containing_numeric_char',
    'number_of_numeric_values_equal_-1', 'most_common_numeric_tokens', 'most_common_tokens',
    'ratio_of_distinct_tokens', 'number_of_missing_values',
    'number_of_distinct_tokens_split_by_punctuation', 'number_of_distinct_tokens',
    'ratio_of_missing_values', 'semantic_types', 'number_of_numeric_values_equal_0',
    'number_of_positive_numeric_values', 'most_common_alphanumeric_tokens',
    'numeric_char_density', 'ratio_of_distinct_values', 'number_of_negative_numeric_values',
    'target_values', 'ratio_of_tokens_split_by_punctuation_containing_numeric_char',
    'ratio_of_values_with_leading_spaces', 'number_of_values_with_trailing_spaces',
    'ratio_of_values_with_trailing_spaces', 'number_of_numeric_values_equal_1',
    'natural_language_of_feature', 'most_common_punctuations', 'spearman_correlation_of_features',
    'number_of_values_with_leading_spaces', 'ratio_of_tokens_containing_numeric_char',
    'number_of_tokens_split_by_punctuation_containing_numeric_char', 'number_of_numeric_values',
    'ratio_of_distinct_tokens_split_by_punctuation', 'number_of_values_containing_numeric_char',
    'most_common_tokens_split_by_punctuation', 'number_of_distinct_values',
    'pearson_correlation_of_features']

default_metafeatures = [
    'ratio_of_values_containing_numeric_char', 'ratio_of_numeric_values',
    'number_of_outlier_numeric_values', 'num_filename', 'number_of_tokens_containing_numeric_char']


class DSboxProfiler(object):
    """
    data profiler moduel. Now only supports csv data.

    Parameters:
    ----------
    _punctuation_outlier_weight: a integer
        the coefficient used in outlier detection for punctuation. default is 3

    _numerical_outlier_weight

    _token_delimiter: a string
        delimiter that used to seperate tokens, default is blank space " ".

    _detect_language: boolean
        true: do detect language; false: not detect language

    _topk: a integer

    Attributes:
    ----------
    """

    def __init__(self, compute_features=None) -> None:

        # All other attributes must be private with leading underscore
        self._punctuation_outlier_weight = 3
        self._numerical_outlier_weight = 3
        self._token_delimiter = " "
        self._detect_language = False
        self._topk = 10
        # list of specified features to compute
        self._specified_features = compute_features if compute_features else default_metafeatures

    def profile(self, inputs: pd.DataFrame, metadata: dict, selected_columns: set = None) -> dict:
        """Save metadata json to file.

        Args:
            inputs: pandas dataframe
            metadata: dict

        Returns:
            dict
        """

        metadata = self._profile_data(inputs, metadata, selected_columns)

        return metadata

    def _profile_data(self, data: pd.DataFrame, metadata: dict, selected_columns: set = None) -> dict:

        """Save metadata json to file.

        Args:
            data: pandas dataframe
            metadata: dict

        Returns:
            dict with dsbox profiled fields
        """

        # STEP 1: data-level calculations
        if "pearson_correlation_of_features" in self._specified_features:
            corr_pearson = data.corr()
            corr_columns = list(corr_pearson.columns)

        if "spearman_correlation_of_features" in self._specified_features:
            corr_spearman = data.corr(method='spearman')
            corr_columns = list(corr_spearman.columns)

        # STEP 2: column-level calculations
        column_counter = -1
        for column_name in data:
            column_counter += 1
            if selected_columns and column_counter not in selected_columns:
                continue
            col = data[column_name]
            # dict: map feature name to content
            each_res = dict()

            if "spearman_correlation_of_features" in self._specified_features and column_name in corr_columns:
                stats_sp = corr_spearman[column_name].describe()
                each_res["spearman_correlation_of_features"] = {
                    'min': stats_sp['min'],
                    'max': stats_sp['max'],
                    'mean': stats_sp['mean'],
                    'median': stats_sp['50%'],
                    'std': stats_sp['std']
                }

            if "spearman_correlation_of_features" in self._specified_features and column_name in corr_columns:
                stats_pr = corr_pearson[column_name].describe()
                each_res["pearson_correlation_of_features"] = {
                    'min': stats_pr['min'],
                    'max': stats_pr['max'],
                    'mean': stats_pr['mean'],
                    'median': stats_pr['50%'],
                    'std': stats_pr['std']
                }

            if col.dtype.kind in np.typecodes['AllInteger'] + 'uMmf':
                if "number_of_missing_values" in self._specified_features:
                    each_res["number_of_missing_values"] = pd.isnull(col).sum()
                if "ratio_of_missing_values" in self._specified_features:
                    each_res["ratio_of_missing_values"] = pd.isnull(col).sum() / col.size
                if "number_of_distinct_values" in self._specified_features:
                    each_res["number_of_distinct_values"] = col.nunique()
                if "ratio_of_distinct_values" in self._specified_features:
                    each_res["ratio_of_distinct_values"] = col.nunique() / float(col.size)

            if col.dtype.kind == 'b':
                if "most_common_raw_values" in self._specified_features:
                    fc_hih.compute_common_values(col.dropna().astype(str), each_res, self._topk)

            elif col.dtype.kind in np.typecodes['AllInteger'] + 'uf':
                fc_hih.compute_numerics(col, each_res,
                                        self._specified_features)  # TODO: do the checks inside the function
                if "most_common_raw_values" in self._specified_features:
                    fc_hih.compute_common_values(col.dropna().astype(str), each_res, self._topk)

            else:

                # Need to compute str missing values before fillna
                if "number_of_missing_values" in self._specified_features:
                    each_res["number_of_missing_values"] = pd.isnull(col).sum()
                if "ratio_of_missing_values" in self._specified_features:
                    each_res["ratio_of_missing_values"] = pd.isnull(col).sum() / col.size

                col = col.astype(object).fillna('').astype(str)

                # compute_missing_space Must be put as the first one because it may change the data content,
                # see function def for details
                fc_lfh.compute_missing_space(col, each_res, self._specified_features)
                # fc_lfh.compute_filename(col, each_res)
                fc_lfh.compute_length_distinct(col, each_res, delimiter=self._token_delimiter,
                                               feature_list=self._specified_features)
                if "natural_language_of_feature" in self._specified_features:
                    fc_lfh.compute_lang(col, each_res)
                if "most_common_punctuations" in self._specified_features:
                    fc_lfh.compute_punctuation(col, each_res, weight_outlier=self._punctuation_outlier_weight)

                fc_hih.compute_numerics(col, each_res, self._specified_features)
                if "most_common_numeric_tokens" in self._specified_features:
                    fc_hih.compute_common_numeric_tokens(col, each_res, self._topk)
                if "most_common_alphanumeric_tokens" in self._specified_features:
                    fc_hih.compute_common_alphanumeric_tokens(col, each_res, self._topk)
                if "most_common_raw_values" in self._specified_features:
                    fc_hih.compute_common_values(col, each_res, self._topk)
                fc_hih.compute_common_tokens(col, each_res, self._topk, self._specified_features)
                if "numeric_char_density" in self._specified_features:
                    fc_hih.compute_numeric_density(col, each_res)
                fc_hih.compute_contain_numeric_values(col, each_res, self._specified_features)
                fc_hih.compute_common_tokens_by_puncs(col, each_res, self._topk, self._specified_features)

            # update metadata for a specific column
            metadata["variables"][column_counter]["dsbox_profiled"] = each_res

        return metadata



from datamart_isi.joiners.join_feature.feature_base import *
from datamart_isi.joiners.join_feature.wrapped_similarity_functions import *


class CategoricalNumberFeature(FeatureBase):
    pass


class CategoricalStringFeature(FeatureBase):
    def similarity_functions(self):
        return [lambda left, right: 1 if to_lower(left) == to_lower(right) else 0]


class CategoricalTokenFeature(FeatureBase):
    """
    Categorical on tokens of feature values, rather than on feature values, then run set based similarity .

    """

    def value_merge_func(self, record: Record):
        # TODO: if multi-col, compare together or separate?
        tokens = []
        for header in self._headers:
            tokens += getattr(record, header).split()
        return tokens

    def similarity_functions(self):
        return [jaccard_sim, cosine_sim]


class NonCategoricalNumberFeature(FeatureBase):
    """
    Non categorical number feature, compare similarity by how different the two values are .

    """
    def __init__(self, df: pd.DataFrame, indexes, metadata, distribute_type, data_type):
        super().__init__(df, indexes, metadata, distribute_type, data_type)
        # TODO: calc the properties below(or copy from profiler info in metadata)
        self._number_type = NumberType.FLOAT
        self._min_value = 0
        self._max_value = 10
        self._sigma = 0.5
        self._max_minus_min = self._max_value - self._min_value

    @property
    def min_value(self):
        return self._min_value

    @property
    def max_value(self):
        return self._max_value

    @property
    def max_minus_min(self):
        return self._max_minus_min

    @property
    def sigma(self):
        return self._sigma

    def value_merge_func(self, record: Record):
        """
        sum the values
        TODO: how to merge?
        """
        return sum([float(getattr(record, header)) for header in self._headers])

    def similarity_functions(self):
        def similarity_function(left, right):
            # TODO: calc a convert from differences to similarity by the range/variance etc
            if self.max_minus_min:
                diff = abs(left-right)/self.max_minus_min
                return pow((1 - diff), self.sigma)
            else:
                return 0
        return [similarity_function]


class NonCategoricalStringFeature(FeatureBase):
    # TODO: if self.multi_column = True, only use set based similarity
    # TODO: TFIDF in rltk needs more pre-calc info
    function_mapping = {
        StringType.WORD: [levenshtein_sim, jaro_winkler_sim, ngram_sim],
        StringType.PHRASE: [hybrid_jaccard_sim, ngram_sim],
        StringType.SENTENCE: [hybrid_jaccard_sim],
        StringType.PARAGRAPH: [hybrid_jaccard_sim],
        StringType.OTHER: []
    }

    def __init__(self, df: pd.DataFrame, indexes, metadata, distribute_type, data_type):
        super().__init__(df, indexes, metadata, distribute_type, data_type)
        # TODO: calc the properties below(or copy from profiler info in metadata)
        self._string_type = StringType.WORD
        # more profiled info needed ...

    def value_merge_func(self, record: Record):
        return [getattr(record, header).replace('_', ' ') for header in self._headers]

    def similarity_functions(self):
        return NonCategoricalStringFeature.function_mapping[self._string_type]


class DatetimeFeature(FeatureBase):
    """
    TODO
    """

    units = ['year', 'month', 'day', 'minute', 'second', 'ms', 'us', 'ns']

    def __init__(self, df: pd.DataFrame, indexes, metadata, distribute_type, data_type):
        super().__init__(df, indexes, metadata, distribute_type, data_type)
        # TODO: now suppose the column has same resolution, maybe add a profiler for datetime is better
        # TODO: now not recognize the real resolution, use the explicit resolution
        # e.g. date column with "{date} 00:00" -> real resolution should be "day" but now we treat it as "minute"
        self._resolution = self._get_resolution()
        if self.multi_column:
            self._parsed_series = pd.to_datetime(df.iloc[:, indexes])
        else:
            self._parsed_series = pd.to_datetime(df.iloc[:, indexes[0]])
        self._min = self._parsed_series.min()
        self._max = self._parsed_series.max()
        self._range = self._max - self._min

    def value_merge_func(self, record: Record):
        return self._parsed_series.iloc[int(record.id)]

    def similarity_functions(self):
        def compare_time(x, y):
            delta = x - y
            if self._range:
                return 1 - abs(delta/self._range)
            else:
                return 0
        # return [lambda x, y: 1 if to_datetime(x) == to_datetime(y) else 0]
        return [compare_time]

    def _get_resolution(self):
        # datetime_ = pd.to_datetime(self._df.iloc[0, self._indexes])
        return DatetimeResolution.HOUR
